#=========================================================================
# YosysStructuralTranslatorL3.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Provide the yosys-compatible SystemVerilog structural translator."""

from pymtl3.passes.backends.sverilog.errors import SVerilogTranslationError
from pymtl3.passes.backends.sverilog.translation.structural.SVStructuralTranslatorL3 import (
    SVStructuralTranslatorL3,
)
from pymtl3.passes.backends.sverilog.util.utility import make_indent
from pymtl3.passes.rtlir import RTLIRType as rt

from .YosysStructuralTranslatorL2 import YosysStructuralTranslatorL2


class YosysStructuralTranslatorL3(
    YosysStructuralTranslatorL2, SVStructuralTranslatorL3 ):

  #-----------------------------------------------------------------------
  # Helper methods that generate port declarations and connections
  #-----------------------------------------------------------------------

  def ifc_port_gen( s, d, msb, ifc_id, _id, n_dim ):
    template = "{d: <7}logic {packed_type: <8} {id_}"
    if not n_dim:
      id_ = f"{ifc_id}__{_id}"
      packed_type = f"[{msb}:0]"
      return [ template.format( **locals() ) ]
    else:
      ret = []
      for i in range( n_dim[0] ):
        ret += s.ifc_port_gen( d, msb, f"{ifc_id}__{i}", _id, n_dim[1:] )
      return ret

  def ifc_conn_gen( s, d, cpid, _pid, cwid, _wid, idx, n_dim ):
    if d.startswith( "input" ):
      template = "assign {wid}{idx} = {pid};"
    else:
      template = "assign {pid} = {wid}{idx};"

    if not n_dim:
      pid = f"{cpid}__{_pid}"
      wid = f"{cwid}__{_wid}"
      return [ template.format( **locals() ) ]
    else:
      ret = []
      for i in range( n_dim[0] ):
        _cpid = f"{cpid}__{i}"
        _idx  = f"[{i}]{idx}"
        ret += s.ifc_conn_gen( d, _cpid, _pid, cwid, _wid, _idx, n_dim[1:] )
      return ret

  #-----------------------------------------------------------------------
  # Declarations
  #-----------------------------------------------------------------------

  def rtlir_tr_interface_port_decls( s, _port_decls ):
    port_decl_list = sum( _port_decls, [] )
    port_decls, wire_decls, connections = [], [], []
    for dct in port_decl_list:
      port_decls.append( dct["port_decls"] )
      wire_decls.append( dct["wire_decls"] )
      connections.append( dct["connections"] )
    return {
      "port_decls" : port_decls,
      "wire_decls" : wire_decls,
      "connections" : connections
    }

  def rtlir_tr_interface_port_decl( s, m, port_id, port_rtype, port_array_type ):
    def _gen_ifc( id_, ifc, n_dim ):
      if not n_dim:
        ret = []
        all_properties = ifc.get_all_properties_packed()
        for name, _rtype in all_properties:
          if isinstance( _rtype, rt.Array ):
            array_type = _rtype
            rtype = _rtype.get_sub_type()
          else:
            array_type = None
            rtype = _rtype
          ret += s.rtlir_tr_interface_port_decl(
            m, id_+"__"+name, rtype, s.rtlir_tr_unpacked_array_type( array_type ) )
        return ret
      else:
        ret = []
        for i in range( n_dim[0] ):
          ret += _gen_ifc( id_+"__"+str(i), ifc, n_dim[1:] )
        return ret
    if isinstance( port_rtype, rt.Port ):
      _port_dtype = s.rtlir_data_type_translation( m, port_rtype.get_dtype() )
      port_dtype = _port_dtype["raw_dtype"]
      direction = port_rtype.get_direction()
      n_dim = port_array_type["n_dim"]
      port_decl = s.port_gen( direction, port_id, n_dim, port_dtype )
      wire_decl = s.port_wire_gen( port_id, n_dim, port_dtype )
      connections = s.port_connection_gen( direction, port_id, n_dim, port_dtype )
      return [ {
        "port_decls" : port_decl,
        "wire_decls" : wire_decl,
        "connections" : connections
      } ]
    else:
      n_dim = port_array_type["n_dim"]
      return _gen_ifc( port_id, port_rtype, n_dim )

  def rtlir_tr_interface_decls( s, ifc_decls ):
    port_decls, wire_decls, connections = [], [], []
    for ifc_decl in ifc_decls:
      port_decls += ifc_decl["port_decls"]
      wire_decls += ifc_decl["wire_decls"]
      connections += ifc_decl["connections"]
    make_indent( port_decls, 1 )
    make_indent( wire_decls, 1 )
    make_indent( connections, 1 )
    return {
      "port_decls" : ",\n".join( port_decls ),
      "wire_decls" : "\n".join( wire_decls ),
      "connections" : "\n".join( connections ),
    }

  def rtlir_tr_interface_decl( s, ifc_id, ifc_rtype, array_type, ports ):
    ifc_ndim = array_type["n_dim"]
    wire_template = "logic {packed_type: <8} {id_}{array_dim_str};"

    # Assemble the interface port declarations
    ret_port = []
    _port_decls = ports["port_decls"]
    for port_decls in _port_decls:
      for port_decl in port_decls:
        direction = port_decl["direction"]
        msb, id_ = port_decl["msb"], port_decl["id_"]
        ret_port += s.ifc_port_gen( direction, msb, ifc_id, id_, ifc_ndim )

    # Assemble the wire declarations
    ret_wire = []
    _wire_decls = ports["wire_decls"]
    for wire_decls in _wire_decls:
      for wire_decl in wire_decls:
        msb, _id, n_dim = wire_decl["msb"], wire_decl["id_"], wire_decl["n_dim"]
        id_ = ifc_id + "__" + _id
        array_dim_str = s._get_array_dim_str( ifc_ndim + n_dim )
        packed_type = f"[{msb}:0]"
        if n_dim or ifc_ndim or "present" in wire_decl:
          ret_wire.append( wire_template.format( **locals() ) )

    # Assemble the connections
    ret_connections = []
    _connections = ports["connections"]
    for connections in _connections:
      for connection in connections:
        d = connection["direction"]
        pid, wid, idx = connection["pid"], connection["wid"], connection["idx"]
        if idx or ifc_ndim or "present" in connection:
          ret_connections += \
            s.ifc_conn_gen( d, ifc_id, pid, ifc_id, wid, idx, ifc_ndim )

    return {
      "port_decls" : ret_port,
      "wire_decls" : ret_wire,
      "connections" : ret_connections
    }

  #-----------------------------------------------------------------------
  # Signal operations
  #-----------------------------------------------------------------------

  def rtlir_tr_interface_array_index( s, base_signal, index ):
    # Interface array index
    s.deq[-1]['s_index'] += "[{}]"
    s.deq[-1]['index'].append( int(index) )
    return f"{base_signal}[{index}]"

  def rtlir_tr_interface_attr( s, base_signal, attr ):
    # Interface attribute
    s.deq[-1]['s_attr'] += "__{}"
    s.deq[-1]['attr'].append( attr )
    return f"{base_signal}.{attr}"



#=========================================================================
# ImportedObject_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 2, 2019
"""Test if the imported object works correctly."""


from pymtl3 import SimulationPass
from pymtl3.datatypes import Bits1, Bits32, Bits64, clog2, mk_bits
from pymtl3.dsl import Component, InPort, Interface, OutPort, Placeholder, connect
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test import TestVectorSimulator

from ... import ImportConfigs, ImportPass
from ...util.utility import get_dir


def local_do_test( _m ):
  _m.elaborate()
  ipass = ImportPass()
  _m.config_sverilog_import.fill_missing( _m )
  m = ipass.get_imported_object( _m )
  sim = TestVectorSimulator( m, _m._test_vectors, _m._tv_in, _m._tv_out )
  sim.run_test()

def test_reg( do_test ):
  def tv_in( m, test_vector ):
    m.in_ = Bits32( test_vector[0] )
  def tv_out( m, test_vector ):
    if test_vector[1] != '*':
      assert m.out == Bits32( test_vector[1] )
  class VReg( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.config_sverilog_import = ImportConfigs(
          vl_src = get_dir(__file__)+'VReg.sv',
          port_map = {
            "clk" : "clk",
            "reset" : "reset",
            "in_" : "d",
            "out" : "q",
          }
      )
  a = VReg()
  a._test_vectors = [
    [    1,    '*' ],
    [    2,      1 ],
    [   -1,      2 ],
    [   -2,     -1 ],
    [   42,     -2 ],
    [  -42,     42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_vl_uninit( do_test ):
  # Use a latch to test if verilator has correctly set up
  # the inital signal values
  def tv_in( m, test_vector ):
    m.in_ = Bits32( test_vector[0] )
  def tv_out( m, test_vector ):
    assert m.out == Bits32( test_vector[1] )
  class VUninit( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.config_sverilog_import = ImportConfigs(
          vl_src = get_dir(__file__)+'VUninit.sv',
          port_map = {
            "in_" : "d",
            "out" : "q",
          },
          vl_xinit = 'ones',
      )
  a = VUninit()
  a._test_vectors = [
    [    0, 4294967295 ],
    [    2, 4294967295 ],
    [   42,         42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_reg_external_trace( do_test ):
  class VRegTrace( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.config_sverilog_import = ImportConfigs(
          vl_src = get_dir(__file__)+'VRegTrace.sv',
          vl_include = [ get_dir(__file__) ],
          port_map = {
            "in_" : "d",
            "out" : "q",
          },
          external_trace = True,
      )
  a = VRegTrace()
  a.elaborate()
  ipass = ImportPass()
  a.config_sverilog_import.fill_missing( a )
  a = ipass.get_imported_object( a )
  a.apply( SimulationPass() )

  a.in_ = Bits32(1)
  a.tick()
  assert a.line_trace() == 'q =          0'
  a.in_ = Bits32(2)
  a.tick()
  assert a.line_trace() == 'q =          1'
  a.in_ = Bits32(-1)
  a.tick()
  assert a.line_trace() == 'q =          2'
  a.tick()
  # 0xFFFFFFFF unsigned
  assert a.line_trace() == 'q = 4294967295'

def test_reg_incomplete_portmap( do_test ):
  def tv_in( m, test_vector ):
    m.in_ = Bits32( test_vector[0] )
  def tv_out( m, test_vector ):
    if test_vector[1] != '*':
      assert m.out == Bits32( test_vector[1] )
  class VReg( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.config_sverilog_import = ImportConfigs(
          vl_src = get_dir(__file__)+'VReg.sv',
          port_map = {
            "in_" : "d",
            "out" : "q",
          }
      )
  a = VReg()
  a._test_vectors = [
    [    1,    '*' ],
    [    2,      1 ],
    [   -1,      2 ],
    [   -2,     -1 ],
    [   42,     -2 ],
    [  -42,     42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_adder( do_test ):
  def tv_in( m, test_vector ):
    m.in0 = Bits32( test_vector[0] )
    m.in1 = Bits32( test_vector[1] )
    m.cin = Bits1( test_vector[2] )
  def tv_out( m, test_vector ):
    if test_vector[3] != '*':
      assert m.out == Bits32( test_vector[3] )
    if test_vector[4] != '*':
      assert m.cout == Bits32( test_vector[4] )
  class VAdder( Component ):
    def construct( s ):
      s.in0 = InPort( Bits32 )
      s.in1 = InPort( Bits32 )
      s.cin = InPort( Bits1 )
      s.out = OutPort( Bits32 )
      s.cout = OutPort( Bits1 )
      s.config_sverilog_import = ImportConfigs(vl_src = get_dir(__file__)+'VAdder.sv')
  a = VAdder()
  a._test_vectors = [
    [    1,      1,     1,     3, 0 ],
    [    1,     -1,     0,     0, 1 ],
    [   42,     42,     1,    85, 0 ],
    [   42,    -43,     1,     0, 1 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_normal_queue( do_test ):
  def tv_in( m, tv ):
    m.enq_en = Bits1( tv[0] )
    m.enq_msg = Bits32( tv[1] )
    m.deq_en = Bits1( tv[3] )
  def tv_out( m, tv ):
    if tv[2] != '*':
      assert m.enq_rdy == Bits1( tv[2] )
    if tv[4] != '*':
      assert m.deq_rdy == Bits1( tv[5] )
    if tv[5] != '*':
      assert m.deq_msg == Bits32( tv[4] )
  class VQueue( Placeholder, Component ):
    def construct( s, data_width, num_entries, count_width ):
      s.count   =  OutPort( mk_bits( count_width )  )
      s.deq_en  =  InPort( Bits1  )
      s.deq_rdy = OutPort( Bits1  )
      s.deq_msg = OutPort( mk_bits( data_width ) )
      s.enq_en  =  InPort( Bits1  )
      s.enq_rdy = OutPort( Bits1  )
      s.enq_msg =  InPort( mk_bits( data_width ) )
      s.config_sverilog_import = ImportConfigs(vl_src = get_dir(__file__)+'VQueue.sv')
  num_entries = 1
  q = VQueue(
      data_width = 32,
      num_entries = num_entries,
      count_width = clog2(num_entries+1))
  # q.dump_vcd = True
  test_vector = [
    #   enq                deq
    #   en    msg   rdy    en    msg   rdy
    [    1,    42,    1,    0,     0,    0  ],
    [    0,    43,    0,    1,    42,    1  ],
    [    1,    43,    1,    0,    42,    0  ],
    [    0,    44,    0,    1,    43,    1  ],
    [    1,    44,    1,    0,    43,    0  ],
    [    0,    45,    0,    1,    44,    1  ],
    [    1,    45,    1,    0,    44,    0  ],
  ]
  q._test_vectors = test_vector
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )



#=========================================================================
# BehavioralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 20, 2019
"""Test the level 2 behavioral translator."""

import pytest

from ...testcases import (
    CaseBits32ConstBitsToTmpVarComp,
    CaseBits32ConstIntToTmpVarComp,
    CaseBits32FreeVarToTmpVarComp,
    CaseBits32MultiTmpWireComp,
    CaseBits32TmpWireAliasComp,
    CaseBits32TmpWireComp,
)
from ..BehavioralTranslatorL2 import BehavioralTranslatorL2
from .TestBehavioralTranslator import mk_TestBehavioralTranslator


@pytest.mark.parametrize(
  'case', [
    CaseBits32TmpWireComp,
    CaseBits32TmpWireAliasComp,
    CaseBits32MultiTmpWireComp,
    CaseBits32FreeVarToTmpVarComp,
    CaseBits32ConstBitsToTmpVarComp,
    CaseBits32ConstIntToTmpVarComp,
  ]
)
def test_generic_behavioral_L2( case ):
  m = case.DUT()
  m.elaborate()
  tr = mk_TestBehavioralTranslator(BehavioralTranslatorL2)(m)
  tr.clear( m )
  tr.translate_behavioral( m )
  upblk_src = tr.behavioral.upblk_srcs[m]
  decl_freevars = tr.behavioral.decl_freevars[m]
  decl_tmpvars = tr.behavioral.decl_tmpvars[m]
  assert upblk_src == case.REF_UPBLK
  assert decl_freevars == case.REF_FREEVAR
  assert decl_tmpvars == case.REF_TMPVAR



#=========================================================================
# VcdGenerationPass_test.py
#=========================================================================
# Perform limited tests on the VCD generation pass. These tests are limited
# in the sense that they do not compare the entire output against some
# reference output, which is hard to obtain in the case of VCD generation.
# Our goal is to have some regression test cases that can hopefully inform
# us of any incompatible changes that lead to the failure of VCD generation
# during a major update of PyMTL.
#
# Author: Peitian Pan
# Date:   Nov 1, 2019

from pymtl3.datatypes import *
from pymtl3.dsl import *
from pymtl3.passes import TracingConfigs
from pymtl3.passes.PassGroups import SimulationPass


def run_test( dut, tv, tv_in, tv_out ):
  vcd_file_name = dut.__class__.__name__ + "_funky"
  dut.config_tracing = TracingConfigs( tracing='vcd', vcd_file_name=vcd_file_name )
  dut.elaborate()
  dut.apply( SimulationPass() )
  for v in tv:
    tv_in( dut, v )
    dut.tick()
    tv_out( dut, v )
  with open( vcd_file_name+".vcd", "r" ) as fd:
    file_str = ''.join( fd.readlines() )
    all_signals = dut.get_input_value_ports() | \
                  dut.get_output_value_ports() | \
                  dut.get_wires()
    for signal in all_signals:
      assert signal._dsl.my_name in file_str

def test_vector_signals():
  class A( Component ):
    def construct( s ):
      s.in0 = InPort( Bits32 )
      s.in1 = InPort( Bits32 )
      s.out = OutPort( Bits32 )

      @s.update
      def add_upblk():
        s.out = s.in0 + s.in1
  def tv_in( m, tv ):
    m.in0 = tv[0]
    m.in1 = tv[1]
  def tv_out( m, tv ):
    assert m.out == tv[2]

  run_test( A(), [
    #     in0      in1      out
    [  b32(0), b32(-1), b32(-1), ],
    [  b32(1),  b32(1),  b32(2), ],
    [ b32(-1),  b32(0), b32(-1), ],
    [ b32(-1),  b32(0), b32(-1), ],
    [ b32(-1),  b32(0), b32(-1), ],
    [ b32(-1),  b32(0), b32(-1), ],
    [ b32(42), b32(42), b32(84), ],
  ], tv_in, tv_out )

def test_bitstruct_signals():
  bs = mk_bitstruct( "BitStructType", {
    'foo' : Bits1,
    'bar' : Bits32,
  } )

  class A2( Component ):
    def construct( s ):
      s.in0 = InPort( bs )
      s.in1 = InPort( Bits32 )
      s.out = OutPort( Bits32 )

      @s.update
      def add_upblk():
        s.out = s.in0.bar + s.in1
  def tv_in( m, tv ):
    m.in0 = tv[0]
    m.in1 = tv[1]
  def tv_out( m, tv ):
    assert m.out == tv[2]

  run_test( A2(), [
    #     in0                 in1      out
    [  bs(b1(0), b32(0)), b32(-1), b32(-1), ],
    [  bs(b1(0), b32(1)),  b32(1),  b32(2), ],
    [ bs(b1(0), b32(-1)),  b32(0), b32(-1), ],
    [ bs(b1(0), b32(42)), b32(42), b32(84), ],
  ], tv_in, tv_out )



#=========================================================================
# StructuralTranslatorL2.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 24, 2019
"""Provide L2 structural translator."""

from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import StructuralRTLIRSignalExpr as sexp
from pymtl3.passes.rtlir.structural.StructuralRTLIRGenL2Pass import (
    StructuralRTLIRGenL2Pass,
)

from .StructuralTranslatorL1 import StructuralTranslatorL1


class StructuralTranslatorL2( StructuralTranslatorL1 ):

  def __init__( s, top ):
    super().__init__( top )

  def clear( s, tr_top ):
    super().clear( tr_top )
    # Declarations
    s.structural.decl_type_struct = []

  #-----------------------------------------------------------------------
  # gen_structural_trans_metadata
  #-----------------------------------------------------------------------

  # Override
  def gen_structural_trans_metadata( s, tr_top ):
    tr_top.apply( StructuralRTLIRGenL2Pass( s.inst_conns ) )

  #-----------------------------------------------------------------------
  # translate_structural
  #-----------------------------------------------------------------------

  # Override
  def translate_structural( s, tr_top ):
    super().translate_structural( tr_top )

  #-----------------------------------------------------------------------
  # _translate_structural
  #-----------------------------------------------------------------------

  # Override
  def _translate_structural( s, m ):
    super()._translate_structural( m )

  #-----------------------------------------------------------------------
  # translate_decls
  #-----------------------------------------------------------------------

  # Override
  def translate_decls( s, m ):
    super().translate_decls( m )

  #-----------------------------------------------------------------------
  # rtlir_data_type_translation
  #-----------------------------------------------------------------------

  # Override
  def rtlir_data_type_translation( s, m, dtype ):
    def recurse_struct_dtype_translation( dtype ):
      for key, value in dtype.get_all_properties().items():
        if isinstance( value, rdt.PackedArray ):
          value = value.get_sub_dtype()
        if not isinstance(value, rdt.Struct):
          s.rtlir_data_type_translation( m, value )
        else:
          s.rtlir_data_type_translation( m, value )

    if isinstance( dtype, rdt.Struct ):
      ret = s.rtlir_tr_struct_dtype( dtype )
      ns_struct = s.structural.decl_type_struct
      if all( dtype != x[0] for x in ns_struct ):
        recurse_struct_dtype_translation( dtype )
        s.structural.decl_type_struct.append( ( dtype, ret ) )
      return ret
    else:
      return super(). \
          rtlir_data_type_translation( m, dtype )

  #-----------------------------------------------------------------------
  # rtlir_signal_expr_translation
  #-----------------------------------------------------------------------

  # Override
  def rtlir_signal_expr_translation( s, expr, m ):
    """Translate a signal expression in RTLIR into its backend representation.

    Add support for the following operations at L2: sexp.PackedIndex, sexp.StructAttr
    """
    if isinstance( expr, sexp.PackedIndex ):
      return s.rtlir_tr_packed_index(
        s.rtlir_signal_expr_translation(expr.get_base(), m), expr.get_index())
    elif isinstance( expr, sexp.StructAttr ):
      return s.rtlir_tr_struct_attr(
        s.rtlir_signal_expr_translation(expr.get_base(), m), expr.get_attr())
    elif isinstance( expr, sexp.ConstInstance ):
      rtype = expr.get_rtype()
      value = expr.get_value()
      dtype = rtype.get_dtype()
      if isinstance( dtype, rdt.Struct ):
        return s.rtlir_tr_struct_instance(dtype, value)
      else:
        return super(). \
            rtlir_signal_expr_translation( expr, m )
    else:
      return super(). \
          rtlir_signal_expr_translation( expr, m )

  #-----------------------------------------------------------------------
  # Methods to be implemented by the backend translator
  #-----------------------------------------------------------------------

  # Data types
  def rtlir_tr_struct_dtype( s, Type ):
    raise NotImplementedError()

  # Signal operations
  def rtlir_tr_packed_index( s, base_signal, index ):
    raise NotImplementedError()

  def rtlir_tr_struct_attr( s, base_signal, attr ):
    raise NotImplementedError()

  def rtlir_tr_struct_instance( s, dtype, struct ):
    raise NotImplementedError()



#=========================================================================
# StructuralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 21, 2019
"""Test the level 2 structural translators."""

import pytest

from ...testcases import (
    CaseNestedPackedArrayStructComp,
    CaseNestedStructPortOnly,
    CaseStructConstComp,
    CaseStructPortOnly,
    CaseStructWireDrivenComp,
    CaseStructx5PortOnly,
)
from ..StructuralTranslatorL2 import StructuralTranslatorL2
from .TestStructuralTranslator import mk_TestStructuralTranslator


def run_test( case, m ):
  m.elaborate()
  tr = mk_TestStructuralTranslator(StructuralTranslatorL2)(m)
  tr.clear( m )
  tr.translate_structural(m)
  try:
    name = tr.structural.component_unique_name[m]
    assert name == case.REF_NAME
    decl_ports = tr.structural.decl_ports[m]
    assert decl_ports == case.REF_PORT
    decl_wires = tr.structural.decl_wires[m]
    assert decl_wires == case.REF_WIRE
    decl_consts = tr.structural.decl_consts[m]
    assert decl_consts == case.REF_CONST
    connections = tr.structural.connections[m]
    assert connections == case.REF_CONN
    struct_types = tr.structural.decl_type_struct
    assert sorted(struct_types, key=lambda x: str(x[1])) == case.REF_STRUCT
  except AttributeError:
    pass

@pytest.mark.parametrize(
  'case', [
    CaseStructPortOnly,
    CaseStructWireDrivenComp,
    CaseStructConstComp,
    CaseStructx5PortOnly,
    CaseNestedStructPortOnly,
    CaseNestedPackedArrayStructComp,
  ]
)
def test_generic_structural_L2( case ):
  run_test( case, case.DUT() )



"""
========================================================================
ComponentLevel1.py
========================================================================
At the bottom level, we only have update blocks and explicit constraints
between two update blocks/one update block. Basically this layer defines
the scheduling policy/elaboration process.
Each update block is called exactly once every cycle. PyMTL will
schedule all update blocks based on the constraints. A total constraint
between two update blocks specifies the order of the two blocks, i.e.
call A before B.
We collect one type of explicit constraints at this level:
* Block constraint: s.add_constraints( U(upA) < U(upB) )

Author : Shunning Jiang
Date   : Nov 3, 2018
"""
import re

from .ConstraintTypes import U
from .errors import InvalidPlaceholderError, NotElaboratedError, UpblkFuncSameNameError
from .NamedObject import NamedObject
from .Placeholder import Placeholder

p = re.compile('( *((@|def).*))')

class ComponentLevel1( NamedObject ):

  #-----------------------------------------------------------------------
  # Private methods
  #-----------------------------------------------------------------------

  def __new__( cls, *args, **kwargs ):
    """ Convention: variables local to the object is created in __new__ """

    inst = super().__new__( cls, *args, **kwargs )

    inst._dsl.name_upblk  = {}
    inst._dsl.upblks      = set()
    inst._dsl.upblk_order = []
    inst._dsl.U_U_constraints = set() # contains ( id(func), id(func) )s

    return inst

  def _collect_vars( s, m ):
    """ Called on individual objects during elaboration/addcomponent.
    The general format resembles "s._all_X.update/append( s._X ). """

    if isinstance( m, ComponentLevel1 ):
      s._dsl.all_upblks |= m._dsl.upblks
      for blk in m._dsl.upblks:
        s._dsl.all_upblk_hostobj[ blk ] = m
      s._dsl.all_U_U_constraints |= m._dsl.U_U_constraints

  def _uncollect_vars( s, m ):

    if isinstance( m, ComponentLevel1 ):
      s._dsl.all_upblks -= m._dsl.upblks
      for k in m._dsl.upblks:
        del s._dsl.all_upblk_hostobj[ k ]
      s._dsl.all_U_U_constraints -= m._dsl.U_U_constraints

  #-----------------------------------------------------------------------
  # Construction-time APIs
  #-----------------------------------------------------------------------

  def update( s, blk ):
    if isinstance( s, Placeholder ):
      raise InvalidPlaceholderError( "Cannot define update block <{}> "
              "in a placeholder component.".format( blk.__name__ ) )
    name = blk.__name__
    if name in s._dsl.name_upblk:
      raise UpblkFuncSameNameError( name )

    s._dsl.name_upblk[ name ] = blk
    s._dsl.upblks.add( blk )
    s._dsl.upblk_order.append( blk )
    return blk

  def get_update_block( s, name ):
    return s._dsl.name_upblk[ name ]

  def add_constraints( s, *args ):
    if isinstance( s, Placeholder ):
      raise InvalidPlaceholderError( "Cannot define constraints "
              "in a placeholder component.".format( blk.__name__ ) )
    for (x0, x1, is_equal) in args:
      assert is_equal == False
      assert isinstance( x0, U ) and isinstance( x1, U ), "Only accept up1<up2"
      assert (x0.func, x1.func) not in s._dsl.U_U_constraints, \
        "Duplicated constraint"
      s._dsl.U_U_constraints.add( (x0.func, x1.func) )

  def construct( s, *args, **kwargs ):
    raise NotImplementedError("construct method, where the design is built,"
                              " is not implemented in {}".format( s.__class__.__name__ ) )

  #-----------------------------------------------------------------------
  # elaborate
  #-----------------------------------------------------------------------

  # We refactor the monolithic elaborate function into smaller functions

  def _elaborate_declare_vars( s ):
    """ Convention: the top level component on which we call elaborate
    declare variables in _declare_vars; it shouldn't have them before
    elaboration.

    Convention: the variables that hold all metadata of descendants
    should have _all prefix."""

    s._dsl.all_upblks = set()
    s._dsl.all_upblk_hostobj = {}
    s._dsl.all_U_U_constraints = set()

    # Although this all_components is a subset of all_named_objects in
    # NamedObject class, I still maintain it here because we want to avoid
    # redundant isinstance check. This being said, I'm going to pay the
    # extra cost of removing from both all_named_objects and this
    # all_components when I delete a component
    s._dsl.all_components = set()

  def _elaborate_collect_all_vars( s ):
    for c in s._dsl.all_named_objects:
      if isinstance( c, ComponentLevel1 ):
        s._dsl.all_components.add( c )
        s._collect_vars( c )

  def elaborate( s ):
    # Directly use the base class elaborate
    NamedObject.elaborate( s )

    s._elaborate_declare_vars()
    s._elaborate_collect_all_vars()

  #-----------------------------------------------------------------------
  # Post-elaborate public APIs (can only be called after elaboration)
  #-----------------------------------------------------------------------

  def is_component( s ):
    return True

  def is_signal( s ):
    return False

  def is_interface( s ):
    return False



#=========================================================================
# DynamicSchedulePass_test.py
#=========================================================================
#
# Author : Shunning Jiang
# Date   : Apr 19, 2019

from pymtl3.datatypes import Bits8, Bits32, bitstruct
from pymtl3.dsl import *
from pymtl3.dsl.errors import UpblkCyclicError

from ..DynamicSchedulePass import DynamicSchedulePass
from ..GenDAGPass import GenDAGPass
from ..SimpleSchedulePass import SimpleSchedulePass
from ..SimpleTickPass import SimpleTickPass


def _test_model( cls ):
  A = cls()
  A.elaborate()
  A.apply( GenDAGPass() )
  A.apply( DynamicSchedulePass() )
  A.apply( SimpleTickPass() )
  A.lock_in_simulation()
  A.eval_combinational()

  T = 0
  while T < 5:
    A.tick()
    print(A.line_trace())
    T += 1
  return A

def test_false_cyclic_dependency():

  class Top(Component):

    def construct( s ):
      s.a = Wire(int)
      s.b = Wire(int)
      s.c = Wire(int)
      s.d = Wire(int)
      s.e = Wire(int)
      s.f = Wire(int)
      s.g = Wire(int)
      s.h = Wire(int)
      s.i = Wire(int)
      s.j = Wire(int)

      @s.update
      def up1():
        s.a = 10 + s.i
        s.b = s.d + 1

      @s.update
      def up2():
        s.c = s.a + 1
        s.e = s.d + 1

      @s.update
      def up3():
        s.d = s.c + 1
        print("up3 prints out d =", s.d)

      @s.update
      def up4():
        s.f = s.d + 1

      @s.update
      def up5():
        s.g = s.c + 1
        s.h = s.j + 1
        print("up5 prints out h =", s.h)

      @s.update
      def up6():
        s.i = s.i + 1

      @s.update
      def up7():
        s.j = s.g + 1

    def done( s ):
      return True

    def line_trace( s ):
      return "a {} | b {} | c {} | d {} | e {} | f {} | g {} | h {} | i {} | j {}" \
              .format( s.a, s.b, s.c, s.d, s.e, s.f, s.g, s.h, s.i, s.j )

  _test_model( Top )

def test_combinational_loop():

  class Top(Component):

    def construct( s ):
      s.a = Wire(int)
      s.b = Wire(int)
      s.c = Wire(int)
      s.d = Wire(int)

      @s.update
      def up1():
        s.b = s.d + 1

      @s.update
      def up2():
        s.c = s.b + 1

      @s.update
      def up3():
        s.d = s.c + 1
        print("up3 prints out d =", s.d)

    def done( s ):
      return True

    def line_trace( s ):
      return "a {} | b {} | c {} | d {}" \
              .format( s.a, s.b, s.c, s.d )

  try:
    _test_model( Top )
  except UpblkCyclicError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown UpblkCyclicError.")

def test_very_deep_dag():

  class Inner(Component):
    def construct( s ):
      s.in_ = InPort(int)
      s.out = OutPort(int)

      @s.update
      def up():
        s.out = s.in_ + 1

    def done( s ):
      return True

    def line_trace( s ):
      return "{} > {}".format( s.a, s.b, s.c, s.d )

  class Top(Component):
    def construct( s, N=2000 ):
      s.inners = [ Inner() for i in range(N) ]
      for i in range(N-1):
        s.inners[i].out //= s.inners[i+1].in_

    def done( s ):
      return True
    def line_trace( s ):
      return ""

  _test_model( Top )

def test_sequential_break_loop():

  class Top(Component):

    def construct( s ):
      s.b = Wire( Bits32 )
      s.c = Wire( Bits32 )

      @s.update
      def up1():
        s.b = s.c + 1

      @s.update_ff
      def up2():
        if s.reset:
          s.c <<= 0
        else:
          s.c <<= s.b + 1

    def done( s ):
      return True

    def line_trace( s ):
      return "b {} | c {}" \
              .format( s.b, s.c )

  A = _test_model( Top )
  assert A.c > 5, "Is the sequential behavior actually captured?"

def test_connect_slice_int():

  class Top( Component ):
    def construct( s ):
      from pymtl3.datatypes import Bits8, Bits32
      s.y = OutPort( Bits8 )
      s.x = Wire( Bits32 )

      s.y //= s.x[0:8]
      @s.update
      def sx():
        s.x = 10 # Except

  try:
    _test_model( Top )
  except TypeError as e:
    assert str(e).startswith( "'int' object is not subscriptable" )
    return
  raise Exception("Should've thrown TypeError: 'int' object is not subscriptable")

def test_const_connect_nested_struct_signal_to_struct():

  @bitstruct
  class SomeMsg1:
    a: Bits8
    b: Bits32

  @bitstruct
  class SomeMsg2:
    a: SomeMsg1
    b: Bits32

  class Top( Component ):
    def construct( s ):
      s.out = OutPort(SomeMsg2)
      connect( s.out, SomeMsg2(SomeMsg1(1,2),3) )

  x = Top()
  x.elaborate()
  x.apply( GenDAGPass() )
  x.apply( DynamicSchedulePass() )
  x.apply( SimpleTickPass() )
  x.lock_in_simulation()
  x.tick()
  assert x.out == SomeMsg2(SomeMsg1(1,2),3)

def test_const_connect_cannot_handle_same_name_nested_struct():

  class A:
    @bitstruct
    class SomeMsg1:
      a: Bits8
      b: Bits32

  class B:
    @bitstruct
    class SomeMsg1:
      c: Bits8
      d: Bits32

  @bitstruct
  class SomeMsg2:
    a: A.SomeMsg1
    b: B.SomeMsg1

  class Top( Component ):
    def construct( s ):
      s.out = OutPort(SomeMsg2)
      connect( s.out, SomeMsg2(A.SomeMsg1(1,2),B.SomeMsg1(3,4)) )

  x = Top()
  x.elaborate()
  try:
    x.apply( GenDAGPass() )
  except AssertionError as e:
    print(e)
    assert str(e) == "Cannot handle two subfields with the same struct name but different structs"
    return
  raise Exception("Should've thrown AssertionError")



#=========================================================================
# TranslationImport_closed_loop_component_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 6, 2019
"""Closed-loop test cases for translation-import with component."""

import hypothesis.strategies as st
import pytest
from hypothesis import HealthCheck, given, reproduce_failure, settings

from pymtl3.datatypes import Bits1, Bits16, Bits32, bitstruct, clog2, mk_bits
from pymtl3.dsl import Component, InPort, Interface, OutPort, Wire, connect
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..util.test_utility import closed_loop_component_test

too_slow = [ HealthCheck.too_slow ]

def local_do_test( m ):
  closed_loop_component_test( m, m._data )

# Use @given(st.data()) to draw input vector inside the test function
#  - also note that data should the rightmost argument of the test function
# Set deadline to None to avoid checking how long each test spin is
# Set max_examples to limit the number of attempts after multiple successes
# Suppress `too_slow` healthcheck to avoid marking a long test as failed
@given(st.data())
@settings(deadline = None, max_examples = 5, suppress_health_check = too_slow)
@pytest.mark.parametrize("Type", [Bits16, Bits32])
def test_adder( do_test, Type, data ):
  class A( Component ):
    def construct( s, Type ):
      s.in_1 = InPort( Type )
      s.in_2 = InPort( Type )
      s.out = OutPort( Type )
      @s.update
      def add_upblk():
        s.out = s.in_1 + s.in_2
    def line_trace( s ): return "sum = " + str( s.out )
  a = A( Type )
  a._data = data
  do_test( a )

@given(st.data())
@settings(deadline = None, max_examples = 5, suppress_health_check = too_slow)
@pytest.mark.parametrize(
  "Type, n_ports", [ (Bits16, 2), (Bits16, 4), (Bits32, 2), (Bits32, 4) ]
)
def test_mux( do_test, Type, n_ports, data ):
  class A( Component ):
    def construct( s, Type, n_ports ):
      s.in_ = [ InPort( Type ) for _ in range(n_ports) ]
      s.sel = InPort( mk_bits( clog2(n_ports) ) )
      s.out = OutPort( Type )
      @s.update
      def add_upblk():
        s.out = s.in_[ s.sel ]
    def line_trace( s ): return "out = " + str( s.out )
  a = A( Type, n_ports )
  a._data = data
  do_test( a )

@given(st.data())
@settings(deadline = None, max_examples = 5, suppress_health_check = too_slow)
def test_struct( do_test, data ):
  @bitstruct
  class strc:
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( strc )
      s.out = OutPort( Bits32 )
      connect( s.out, s.in_.foo )
    def line_trace( s ): return "out = " + str( s.out )
  a = A()
  a._data = data
  do_test( a )

@given(st.data())
@settings(deadline = None, max_examples = 10, suppress_health_check = too_slow)
def test_nested_struct( do_test, data ):
  @bitstruct
  class inner_struct:
    bar: Bits32
  @bitstruct
  class strc:
    foo: Bits32
    inner: inner_struct
    packed_array: [ [ Bits16 ]*2 ] *3
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( strc )
      s.out_foo = OutPort( Bits32 )
      s.out_bar = OutPort( Bits32 )
      s.out_sum = OutPort( Bits16 )
      s.sum = [ Wire( Bits16 ) for _ in range(3) ]
      @s.update
      def upblk():
        for i in range(3):
          s.sum[i] = s.in_.packed_array[i][0] + s.in_.packed_array[i][1]
        s.out_sum = s.sum[0] + s.sum[1] + s.sum[2]
      connect( s.out_foo, s.in_.foo )
      connect( s.out_bar, s.in_.inner.bar )
    def line_trace( s ): return "out_sum = " + str( s.out_sum )
  a = A()
  a._data = data
  do_test( a )

@given(st.data())
@settings(deadline = None, max_examples = 10, suppress_health_check = too_slow)
def test_subcomp( do_test, data ):
  @bitstruct
  class inner_struct:
    bar: Bits32
  @bitstruct
  class strc:
    foo: Bits32
    inner: inner_struct
    packed_array: [ [ Bits16 ]*2 ] *3
  class B( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      connect( s.out, 0 )
  class A( Component ):
    def construct( s ):
      s.b = B()
      s.in_ = InPort( strc )
      s.out_foo = OutPort( Bits32 )
      s.out_bar = OutPort( Bits32 )
      s.out_sum = OutPort( Bits16 )
      s.sum = [ Wire( Bits16 ) for _ in range(3) ]
      @s.update
      def upblk():
        for i in range(3):
          s.sum[i] = s.in_.packed_array[i][0] + s.in_.packed_array[i][1]
        s.out_sum = s.sum[0] + s.sum[1] + s.sum[2]
      connect( s.out_foo, s.b.out )
      connect( s.out_bar, s.in_.inner.bar )
    def line_trace( s ): return "out_sum = " + str( s.out_sum )
  a = A()
  a._data = data
  do_test( a )

# Test contributed by Cheng Tan
@given( st.data() )
@settings( deadline = None, max_examples = 5 )
@pytest.mark.parametrize( "Type", [ Bits16, Bits32 ] )
def test_index_static( do_test, Type, data ):
  class A( Component ):
    def construct( s, Type ):
      s.in_ = [InPort ( Type ) for _ in range(2)]
      s.out = [OutPort( Type ) for _ in range(2)]
      @s.update
      def index_upblk():
        if s.in_[0] > s.in_[1]:
          s.out[0] = Type(1)
          s.out[1] = Type(0)
        else:
          s.out[0] = Type(0)
          s.out[1] = Type(1)

    def line_trace( s ): return "s.in0  = " + str( s.in_[0] ) +\
                                "s.in1  = " + str( s.in_[1] ) +\
                                "s.out0 = " + str( s.out[0] ) +\
                                "s.out1 = " + str( s.out[1] )
  a = A( Type )
  a._data = data
  do_test( a )



#=========================================================================
# BehavioralRTLIRL4Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 2, 2019
"""Test the level 4 behavioral RTLIR passes.

The L4 generation, L4 type check, and visualization passes are invoked. The
generation pass results are verified against a reference AST.
"""

from pymtl3.dsl.errors import VarNotDeclaredError
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIR import *
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL4Pass import (
    BehavioralRTLIRGenL4Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL4Pass import (
    BehavioralRTLIRTypeCheckL4Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRVisualizationPass import (
    BehavioralRTLIRVisualizationPass,
)
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError, PyMTLTypeError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure
from pymtl3.passes.testcases import (
    CaseArrayInterfacesComp,
    CaseInterfaceAttributeComp,
    CaseInterfaceMissingAttributeComp,
)


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""
  if isinstance(m, type):
    m = m.DUT()
  m.elaborate()
  m.apply( BehavioralRTLIRGenL4Pass() )
  m.apply( BehavioralRTLIRTypeCheckL4Pass() )
  m.apply( BehavioralRTLIRVisualizationPass() )

  try:
    ref = m._rtlir_test_ref
    for blk in m.get_update_blocks():
      upblk = m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ]
      assert upblk == ref[ blk.__name__ ]
  except AttributeError:
    pass

#-------------------------------------------------------------------------
# Correct test cases
#-------------------------------------------------------------------------

def test_L4_interface_attr( do_test ):
  a = CaseInterfaceAttributeComp.DUT()
  a._rtlir_test_ref = { 'upblk' : CombUpblk( 'upblk', [ Assign(
      Attribute( Base( a ), 'out' ), Attribute(
        Attribute( Base( a ), 'in_' ), 'foo' ), True ) ] ) }
  do_test( a )

def test_L4_interface_array_index( do_test ):
  a = CaseArrayInterfacesComp.DUT()
  a._rtlir_test_ref = { 'upblk' : CombUpblk( 'upblk', [ Assign(
      Attribute( Base( a ), 'out' ), Attribute( Index(
        Attribute( Base( a ), 'in_' ), Number(2) ), 'foo' ), True ) ] ) }
  do_test( a )

#-------------------------------------------------------------------------
# PyMTL type errors
#-------------------------------------------------------------------------

def test_L4_interface_no_field( do_test ):
  with expected_failure( VarNotDeclaredError, 's.in_ does not have field "bar"' ):
    do_test( CaseInterfaceMissingAttributeComp )



"""
========================================================================
WrapGreenletPass.py
========================================================================
Wrap all update blocks that call methods with blocking decorator with
greenlet.

Author : Shunning Jiang
Date   : May 20, 2019
"""
from greenlet import greenlet

from pymtl3.dsl.errors import UpblkCyclicError
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.errors import PassOrderError


class WrapGreenletPass( BasePass ):
  def __call__( self, top ):
    if not hasattr( top, "_dag" ):
      raise PassOrderError( "_dag" )

    self.wrap_greenlet( top )

  def wrap_greenlet( self, top ):

    all_upblks      = top._dag.final_upblks
    all_constraints = top._dag.all_constraints
    greenlet_upblks = top._dag.greenlet_upblks

    if not greenlet_upblks:
      return

    def wrap_greenlet( blk ):

      def greenlet_wrapper():
        while True:
          blk()
          greenlet.getcurrent().parent.switch()

      gl = greenlet( greenlet_wrapper )

      def greenlet_ticker():
        gl.switch()

      # greenlet_ticker.greenlet = gl
      greenlet_ticker.__name__ = blk.__name__

      return greenlet_ticker

    new_upblks  = set()
    wrapped_blk_mapping = {}

    for blk in all_upblks:
      if blk in greenlet_upblks:
        wrapped = wrap_greenlet( blk )
        wrapped_blk_mapping[ blk ] = wrapped
        new_upblks.add( wrapped )
      else:
        new_upblks.add( blk )

    new_constraints = set()

    for (x, y) in all_constraints:
      if x in greenlet_upblks:
        x = wrapped_blk_mapping[ x ]
      if y in greenlet_upblks:
        y = wrapped_blk_mapping[ y ]

      new_constraints.add( (x, y) )

    top._dag.final_upblks    = new_upblks
    top._dag.all_constraints = new_constraints



"""
========================================================================
MemMsg.py
========================================================================
New memory message type implementation.

Author : Shunning Jiang, Yanghui Ou
Date   : Mar 12, 2018
"""
from pymtl3 import *


def mk_mem_msg( opq, addr, data ):
  return mk_mem_req_msg( opq, addr, data ), mk_mem_resp_msg( opq, data )

def mk_mem_req_msg( opq, addr, data ):
  OpqType  = mk_bits( opq            )
  AddrType = mk_bits( addr           )
  LenType  = mk_bits( clog2(data>>3) )
  DataType = mk_bits( data           )
  cls_name = "MemReqMsg_{}_{}_{}".format( opq, addr, data )

  def req_to_str( self ):
    return "{}:{}:{}:{}:{}".format(
      MemMsgType.str[ int( self.type_ ) ],
      OpqType( self.opaque ),
      AddrType( self.addr ),
      LenType( self.len ),
      DataType( self.data ) if self.type_ != MemMsgType.READ else
      " " * ( data//4 ),
    )

  req_cls = mk_bitstruct( cls_name, {
    'type_':  Bits4,
    'opaque': OpqType,
    'addr':   AddrType,
    'len':    LenType,
    'data':   DataType,
  },
  namespace = {
    '__str__' : req_to_str
  })

  req_cls.data_nbits = data
  return req_cls

def mk_mem_resp_msg( opq, data ):
  OpqType  = mk_bits( opq            )
  LenType  = mk_bits( clog2(data>>3) )
  DataType = mk_bits( data           )
  cls_name = "MemRespMsg_{}_{}".format( opq, data )

  def resp_to_str( self ):
    return "{}:{}:{}:{}:{}".format(
      MemMsgType.str[ int( self.type_ ) ],
      OpqType( self.opaque ),
      Bits2( self.test ),
      LenType( self.len ),
      DataType( self.data ) if self.type_ != MemMsgType.WRITE else
      " " * ( data//4 ),
    )

  resp_cls = mk_bitstruct( cls_name, {
    'type_':  Bits4,
    'opaque': OpqType,
    'test':   Bits2,
    'len':    LenType,
    'data':   DataType,
  },
  namespace = {
    '__str__' : resp_to_str
  })

  resp_cls.data_nbits = data
  return resp_cls

class MemMsgType:
  READ       = 0
  WRITE      = 1
  # write no-refill
  WRITE_INIT = 2
  AMO_ADD    = 3
  AMO_AND    = 4
  AMO_OR     = 5
  AMO_SWAP   = 6
  AMO_MIN    = 7
  AMO_MINU   = 8
  AMO_MAX    = 9
  AMO_MAXU   = 10
  AMO_XOR    = 11

  str = {
    READ       : "rd",
    WRITE      : "wr",
    WRITE_INIT : "in",
    AMO_ADD    : "ad",
    AMO_AND    : "an",
    AMO_OR     : "or",
    AMO_SWAP   : "sw",
    AMO_MIN    : "mi",
    AMO_MINU   : "mu",
    AMO_MAX    : "mx",
    AMO_MAXU   : "xu",
    AMO_XOR    : "xo",
  }



#=========================================================================
# SVBehavioralTranslatorL3.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 18, 2019
"""Provide the level 3 SystemVerilog translator implementation."""

from pymtl3.passes.backends.generic.behavioral.BehavioralTranslatorL3 import (
    BehavioralTranslatorL3,
)
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt

from ...errors import SVerilogTranslationError
from .SVBehavioralTranslatorL2 import (
    BehavioralRTLIRToSVVisitorL2,
    SVBehavioralTranslatorL2,
)


class SVBehavioralTranslatorL3(
    SVBehavioralTranslatorL2, BehavioralTranslatorL3 ):

  def _get_rtlir2sv_visitor( s ):
    return BehavioralRTLIRToSVVisitorL3

  def rtlir_tr_behavioral_freevar( s, id_, rtype, array_type, dtype, obj ):
    assert isinstance( rtype, rt.Const ), \
      f'{id_} freevar should be a constant!'
    if isinstance( rtype.get_dtype(), rdt.Struct ):
      return s.rtlir_tr_const_decl( f'__const__{id_}', rtype, array_type, dtype, obj )
    else:
      return super().rtlir_tr_behavioral_freevar(
          id_, rtype, array_type, dtype, obj )

#-------------------------------------------------------------------------
# BehavioralRTLIRToSVVisitorL3
#-------------------------------------------------------------------------

class BehavioralRTLIRToSVVisitorL3( BehavioralRTLIRToSVVisitorL2 ):
  """Visitor that translates RTLIR to SystemVerilog for a single upblk."""

  #-----------------------------------------------------------------------
  # visit_StructInst
  #-----------------------------------------------------------------------

  def visit_StructInst( s, node ):
    values = list( map( s.visit, node.values ) )
    if len( values ) == 1:
      return values[0]
    else:
      cat_value = ", ".join( values )
      return f"{{ {cat_value} }}"

  #-----------------------------------------------------------------------
  # visit_Attribute
  #-----------------------------------------------------------------------

  def visit_Attribute( s, node ):
    """Return the SystemVerilog representation of an attribute.

    Add support for accessing struct attribute in L3.
    """
    if isinstance( node.value.Type, rt.Signal ):
      # if isinstance( node.value.Type, rt.Const ):
        # raise SVerilogTranslationError( s.blk, node,
          # "attribute ({}) of constant struct instance ({}) is not supported!". \
            # format( node.attr, node.value ))

      if isinstance( node.value.Type.get_dtype(), rdt.Struct ):
        value = s.visit( node.value )
        attr = node.attr
        s.check_res( node, attr )
        return f'{value}.{attr}'

    return super().visit_Attribute( node )



#=========================================================================
# TranslationImport_closed_loop_component_input_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 6, 2019
"""Closed-loop test cases for translation-import with component and input."""

from random import seed

from pymtl3.passes.backends.sverilog.test.TranslationImport_closed_loop_component_input_test import (
    test_adder,
    test_mux,
)
from pymtl3.passes.backends.sverilog.util.test_utility import (
    closed_loop_component_input_test,
)
from pymtl3.passes.rtlir.util.test_utility import do_test

seed( 0xdeadebeef )

def local_do_test( m ):
  test_vector = m._test_vector
  tv_in = m._tv_in
  closed_loop_component_input_test( m, test_vector, tv_in, "yosys" )



"""
========================================================================
ComponentLevel5.py
========================================================================
We allow CallerPort to be connected to CalleePort

Author : Shunning Jiang
Date   : Dec 29, 2018
"""
from .ComponentLevel1 import ComponentLevel1
from .ComponentLevel2 import ComponentLevel2
from .ComponentLevel4 import ComponentLevel4
from .Connectable import CalleePort, CallerPort, Const, Interface, MethodPort, Signal
from .errors import InvalidConnectionError, MultiWriterError
from .NamedObject import NamedObject
from .Placeholder import Placeholder


# This method_port is a syntactic sugar to create a CalleePort
# Note that for a simple method port we currently don't care about type
def method_port( method ):
  method._callee_port = True
  return method

class ComponentLevel5( ComponentLevel4 ):

  #-----------------------------------------------------------------------
  # Private methods
  #-----------------------------------------------------------------------

  def _handle_decorated_methods( s ):

    for x in s.__class__.__dict__:
      method = getattr( s, x )
      # We identify decorated method port here
      if hasattr( method, "_callee_port" ):
        setattr( s, x, CalleePort( method = method ) )

  # Override
  def _construct( s ):
    """ We override _construct here to add method binding. Basically
    we do this after the class is constructed but before the construct()
    elaboration happens."""

    if not s._dsl.constructed:

      # Merge the actual keyword args and those args set by set_parameter
      if s._dsl.param_tree is None:
        kwargs = s._dsl.kwargs
      elif s._dsl.param_tree.leaf is None:
        kwargs = s._dsl.kwargs
      else:
        kwargs = s._dsl.kwargs.copy()
        if "construct" in s._dsl.param_tree.leaf:
          more_args = s._dsl.param_tree.leaf[ "construct" ]
          kwargs.update( more_args )

      s._handle_decorated_methods()

      # Same as parent class _construct
      s.construct( *s._dsl.args, **kwargs )

      if s._dsl.call_kwargs is not None: # s.a = A()( b = s.b )
        s._continue_call_connect()

      s._dsl.constructed = True

  def _connect_method_ports( s, o1, o2 ):
    s._dsl.adjacency[o1].add( o2 )
    s._dsl.adjacency[o2].add( o1 )
    s._dsl.connect_order.append( (o1, o2) )

  # Override
  def _connect_dispatch( s, o1, o2, o1_connectable, o2_connectable ):

    if o1_connectable and o2_connectable:
      # if both connectable, dispatch signal-signal and interface-interface
      if isinstance( o1, Signal ) and isinstance(o2, Signal ):
        s._connect_signal_signal( o1, o2 )
      elif isinstance( o1, Interface ) and isinstance( o2, Interface ):
        s._connect_interfaces( o1, o2 )
      # Methodport added here
      elif isinstance( o1, MethodPort ) and isinstance( o2, MethodPort ):
        s._connect_method_ports( o1, o2 )
      else:
        raise InvalidConnectionError("{} cannot be connected to {}: {} != {}" \
              .format(repr(o1), repr(o2), type(o1), type(o2)) )
    else:
      # One is connectable, we make sure it's o1
      if o2_connectable:
        o1, o2 = o2, o1
      assert isinstance( o1, Signal ), f"Cannot connect {o2} to {o1!r} of {type(o1)}."

      s._connect_signal_const( o1, o2 )

  def _resolve_method_connections( s ):

    # First of all, bfs the "forest" to find out all nets

    nets = s._floodfill_nets( s._dsl.all_method_ports, s._dsl.all_adjacency )

    # All CalleePort are "writers" because they have actual methods

    ret = []

    for net in nets:
      writer = None

      for member in net:

        if isinstance( member, CalleePort ):
          if member.method is not None or \
             isinstance( member.get_host_component(), Placeholder ):
            if writer is None:
              writer = member
            else:
              raise MultiWriterError( \
                "Two-method conflict \"{}\", \"{}\" in the following net:\n - {}".format(
                repr(member), repr(writer),
                "\n - ".join([repr(x) for x in net])) )

        else:
          assert isinstance( member, CallerPort ), "We don't allow connecting method " \
                                                   "port to other ports of {} type".format( member.__class__ )
      assert writer is not None, "This method net has no actual method to call.\n- {}" \
                                  .format( '\n- '.join([ repr(x) for x in net]) )
      ret.append( (writer, net) )

    return ret

  # TODO Check if all method net port directions are correct

  #-----------------------------------------------------------------------
  # elaborate
  #-----------------------------------------------------------------------
  # We still reuse the elaborate template by adding functionalities to
  # sub-functions called by elaborate

  # Override
  def _elaborate_declare_vars( s ):
    super()._elaborate_declare_vars()

    s._dsl.all_method_ports = set()

  # However, we need to override the whole function here because we want
  # to add some fine-grained functionalities to avoid reduntant isinstance
  # Override
  def _elaborate_collect_all_vars( s ):
    for c in s._dsl.all_named_objects:
      if isinstance( c, Signal ):
        s._dsl.all_signals.add( c )
      elif isinstance( c, ComponentLevel1 ):
        s._dsl.all_components.add( c )
        s._collect_vars( c )
      # Added here
      elif isinstance( c, MethodPort ):
        s._dsl.all_method_ports.add( c )

    s._dsl.all_value_nets  = s._resolve_value_connections()
    # Added here
    s._dsl.all_method_nets = s._resolve_method_connections()
    s._dsl._has_pending_value_connections = False
    s._dsl._has_pending_method_connections = False



#=========================================================================
# BehavioralRTLIRGenL1Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Oct 20, 2018
"""Provide L1 behavioral RTLIR generation pass."""

import ast
import copy

import pymtl3.dsl as dsl
from pymtl3.datatypes import Bits, concat, reduce_and, reduce_or, reduce_xor, sext, zext
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError
from pymtl3.passes.rtlir.util.utility import get_ordered_upblks, get_ordered_update_ff

from . import BehavioralRTLIR as bir


class BehavioralRTLIRGenL1Pass( BasePass ):
  def __call__( s, m ):
    """Generate RTLIR for all upblks of m."""
    if not hasattr( m, '_pass_behavioral_rtlir_gen' ):
      m._pass_behavioral_rtlir_gen = PassMetadata()

    m._pass_behavioral_rtlir_gen.rtlir_upblks = {}
    visitor = BehavioralRTLIRGeneratorL1( m )
    upblks = {
      'CombUpblk' : get_ordered_upblks(m),
      'SeqUpblk'  : get_ordered_update_ff(m),
    }
    # Sort the upblks by their name
    upblks['CombUpblk'].sort( key = lambda x: x.__name__ )
    upblks['SeqUpblk'].sort( key = lambda x: x.__name__ )

    for upblk_type in ( 'CombUpblk', 'SeqUpblk' ):
      for blk in upblks[ upblk_type ]:
        visitor._upblk_type = upblk_type
        upblk_info = m.get_update_block_info( blk )
        upblk = visitor.enter( blk, upblk_info[-1] )
        upblk.is_lambda = upblk_info[0]
        upblk.src       = upblk_info[1]
        upblk.lino      = upblk_info[2]
        upblk.filename  = upblk_info[3]
        m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] = upblk

class BehavioralRTLIRGeneratorL1( ast.NodeVisitor ):
  def __init__( s, component ):
    s.component = component

  def enter( s, blk, ast ):
    """Entry point of RTLIR generation."""
    s.blk     = blk

    # s.globals contains a dict of the global namespace of the module where
    # blk was defined
    s.globals = blk.__globals__

    # s.closure contains the free variables defined in an enclosing scope.
    # Basically this is the model instance s.
    s.closure = {}

    for i, var in enumerate( blk.__code__.co_freevars ):
      try:
        s.closure[ var ] = blk.__closure__[ i ].cell_contents
      except ValueError:
        pass

    ret = s.visit( ast )
    ret.component = s.component
    return ret

  def get_call_obj( s, node ):
    if hasattr(node, "starargs") and node.starargs:
      raise PyMTLSyntaxError( s.blk, node, 'star argument is not supported!')
    if hasattr(node, "kwargs") and node.kwargs:
      raise PyMTLSyntaxError( s.blk, node,
        'double-star argument is not supported!')
    if node.keywords:
      raise PyMTLSyntaxError( s.blk, node, 'keyword argument is not supported!')
    if not isinstance( node.func, ast.Name ):
      raise PyMTLSyntaxError( s.blk, node,
        f'{node.func} is called but is not a name!')
    func = node.func

    # Find the corresponding object of node.func field
    # TODO: Support Verilog task?
    # if func in s.mapping:
      # The node.func field corresponds to a member of this class
      # obj = s.mapping[ func ][ 0 ]
    # else:
    try:
      # An object in global namespace is used
      if func.id in s.globals:
        obj = s.globals[ func.id ]
      # An object in closure is used
      elif func.id in s.closure:
        obj = s.closure[ func.id ]
      else:
        raise NameError
    except NameError:
      raise PyMTLSyntaxError( s.blk, node,
        node.func.id + ' function is not found!' )
    return obj

  def visit_Module( s, node ):
    if len( node.body ) != 1 or \
        not isinstance( node.body[0], ast.FunctionDef ):
      raise PyMTLSyntaxError( s.blk, node,
        'Update blocks should have exactly one FuncDef!' )
    ret = s.visit( node.body[0] )
    ret.ast = node
    return ret

  def visit_FunctionDef( s, node ):
    """Return the behavioral RTLIR of function node.

    We do not need to check the decorator list -- the fact that we are
    visiting this node ensures this node was added to the upblk
    dictionary through s.update() (or other PyMTL decorators) earlier!
    """
    # Check the arguments of the function
    if node.args.args or node.args.vararg or node.args.kwarg:
      raise PyMTLSyntaxError( s.blk, node,
        'Update blocks should not have arguments!' )

    # Save the name of the upblk
    s._upblk_name = node.name

    # Get the type of upblk from ._upblk_type variable
    ret = eval( 'bir.' + s._upblk_type + '( node.name, [] )' )
    for stmt in node.body:
      ret.body.append( s.visit( stmt ) )
    ret.ast = node
    return ret

  def visit_Assign( s, node ):
    if len( node.targets ) != 1:
      raise PyMTLSyntaxError( s.blk, node,
        'Assigning to multiple targets is not allowed!' )

    value = s.visit( node.value )
    target = s.visit( node.targets[0] )
    ret = bir.Assign( target, value, blocking = True )
    ret.ast = node
    return ret

  def visit_AugAssign( s, node ):
    """Return the behavioral RTLIR of a non-blocking assignment

    If the given AugAssign is not non-blocking assignment, throw PyMTLSyntaxError
    """
    if isinstance( node.op, ast.LShift ):
      value = s.visit( node.value )
      target = s.visit( node.target )
      ret = bir.Assign( target, value, blocking = False )
      ret.ast = node
      return ret
    raise PyMTLSyntaxError( s.blk, node,
        'invalid operation: augmented assignment is not non-blocking assignment!' )

  def visit_Call( s, node ):
    """Return the behavioral RTLIR of method calls.

    Some data types are interpreted as function calls in the Python AST.
    Example: Bits4(2)
    These are converted to different RTLIR nodes in different contexts.
    """
    obj = s.get_call_obj( node )
    if ( obj == copy.copy ) or ( obj == copy.deepcopy ):
      if len( node.args ) != 1:
        raise PyMTLSyntaxError( s.blk, node,
          f'copy method {obj} takes exactly 1 argument!')
      ret = s.visit( node.args[0] )
      ret.ast = node
      return ret

    # Now that we have the live Python object, there are a few cases that
    # we need to treat separately:
    # 1. Instantiation: Bits16( 10 ) where obj is an instance of Bits
    # Bits16( 1+2 ), Bits16( s.STATE_A )?
    # 2. concat()
    # 3. zext(), sext()
    # TODO: support the following
    # 4. reduce_and(), reduce_or(), reduce_xor()
    # 5. Real function call: not supported yet

    # Deal with Bits type cast
    if isinstance(obj, type) and issubclass( obj, Bits ):
      nbits = obj.nbits
      if len( node.args ) != 1:
        raise PyMTLSyntaxError( s.blk, node,
          'exactly one argument should be given to Bits!' )
      value = s.visit( node.args[0] )
      ret = bir.SizeCast( nbits, value )
      ret.ast = node
      return ret

    # concat method
    elif obj is concat:
      if len( node.args ) < 1:
        raise PyMTLSyntaxError( s.blk, node,
          'at least one argument should be given to concat!' )
      values = [s.visit(c) for c in node.args]
      ret = bir.Concat( values )
      ret.ast = node
      return ret

    # zext method
    elif obj is zext:
      if len( node.args ) != 2:
        raise PyMTLSyntaxError( s.blk, node,
          'exactly two arguments should be given to zext!' )
      nbits = s.visit( node.args[1] )
      value = s.visit( node.args[0] )
      ret = bir.ZeroExt( nbits, value )
      ret.ast = node
      return ret

    # sext method
    elif obj is sext:
      if len( node.args ) != 2:
        raise PyMTLSyntaxError( s.blk, node,
          'exactly two arguments should be given to sext!' )
      nbits = s.visit( node.args[1] )
      value = s.visit( node.args[0] )
      ret = bir.SignExt( nbits, value )
      ret.ast = node
      return ret

    # reduce methods
    elif obj is reduce_and or obj is reduce_or or obj is reduce_xor:
      if obj is reduce_and:
        op = bir.BitAnd()
      elif obj is reduce_or:
        op = bir.BitOr()
      elif obj is reduce_xor:
        op = bir.BitXor()
      if len( node.args ) != 1:
        raise PyMTLSyntaxError( s.blk, node,
          f'exactly two arguments should be given to reduce {op} methods!' )
      value = s.visit( node.args[0] )
      ret = bir.Reduce( op, value )
      ret.ast = node
      return ret

    else:
      # Only Bits class instantiation is supported at L1
      raise PyMTLSyntaxError( s.blk, node,
        f'Unrecognized method call {obj.__name__}!' )

  def visit_Attribute( s, node ):
    ret = bir.Attribute( s.visit( node.value ), node.attr )
    ret.ast = node
    return ret

  def visit_Subscript( s, node ):
    value = s.visit( node.value )
    if isinstance( node.slice, ast.Slice ):
      if node.slice.step is not None:
        raise PyMTLSyntaxError( s.blk, node,
          'Slice with steps is not supported!' )
      lower, upper = s.visit( node.slice )
      ret = bir.Slice( value, lower, upper )
      ret.ast = node
      return ret

    # signal[ index ]
    # index might be a slice object!
    if isinstance( node.slice, ast.Index ):
      idx = s.visit( node.slice )
      # If we have a static slice object then use it
      if isinstance( idx, bir.FreeVar ) and isinstance( idx.obj, slice ):
        slice_obj = idx.obj
        if slice_obj.step is not None:
          raise PyMTLSyntaxError( s.blk, node,
            'Slice with steps is not supported!' )
        assert isinstance( slice_obj.start, int ) and \
               isinstance( slice_obj.stop, int ), \
            f"start and stop of slice object {slice_obj} must be integers!"
        ret = bir.Slice( value,
              bir.Number(slice_obj.start), bir.Number(slice_obj.stop) )
      # Else this is a real index
      else:
        ret = bir.Index( value, idx )
      ret.ast = node
      return ret

    raise PyMTLSyntaxError( s.blk, node,
      'Illegal subscript ' + node + ' encountered!' )

  def visit_Slice( s, node ):
    return ( s.visit( node.lower ), s.visit( node.upper ) )

  def visit_Index( s, node ):
    return s.visit( node.value )

  def visit_Name( s, node ):
    if node.id in s.closure:
      # free var from closure
      obj = s.closure[ node.id ]
      if isinstance( obj, dsl.Component ):
        # Component freevars are an L1 thing.
        if obj is not s.component:
          raise PyMTLSyntaxError( s.blk, node,
            f'Component {obj} is not a sub-component of {s.component}!' )
        ret = bir.Base( obj )
      else:
        ret =  bir.FreeVar( node.id, obj )
      ret.ast = node
      return ret
    elif node.id in s.globals:
      # free var from the global name space
      ret = bir.FreeVar( node.id, s.globals[ node.id ] )
      ret.ast = node
      return ret
    raise PyMTLSyntaxError( s.blk, node,
      f'Temporary variable {node.id} is not supported at L1!' )

  def visit_Num( s, node ):
    ret = bir.Number( node.n )
    ret.ast = node
    return ret

  def visit_If( s, node ): raise NotImplementedError()

  def visit_For( s, node ): raise NotImplementedError()

  def visit_BoolOp( s, node ): raise NotImplementedError()

  def visit_BinOp( s, node ): raise NotImplementedError()

  def visit_UnaryOp( s, node ): raise NotImplementedError()

  def visit_IfExp( s, node ): raise NotImplementedError()

  def visit_Compare( s, node ): raise NotImplementedError()

  # $display
  def visit_Print( s, node ): raise NotImplementedError()

  # function
  def visit_Return( s, node ): raise NotImplementedError()

  # SV assertion
  def visit_Assert( s, node ): raise NotImplementedError()

  def visit_Expr( s, node ):
    """Return the behavioral RTLIR of an expression.

    ast.Expr might be useful when a statement is only a call to a task or
    a non-returning function.
    """
    raise PyMTLSyntaxError(
      s.blk, node, 'Stand-alone expression is not supported yet!'
    )

  def visit_Lambda( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: lambda function' )

  def visit_Dict( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid type: dict' )

  def visit_Set( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid type: set' )

  def visit_List( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid type: list' )

  def visit_Tuple( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid type: tuple' )

  def visit_ListComp( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: list comprehension' )

  def visit_SetComp( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: set comprehension' )

  def visit_DictComp( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: dict comprehension' )

  def visit_GeneratorExp( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: generator expression' )

  def visit_Yield( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: yield' )

  def visit_Repr( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: repr' )

  def visit_Str( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: str' )

  def visit_ClassDef( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: classdef' )

  def visit_Delete( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: delete' )

  def visit_With( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: with' )

  def visit_Raise( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: raise' )

  def visit_TryExcept( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: try-except' )

  def visit_TryFinally( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: try-finally' )

  def visit_Import( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: import' )

  def visit_ImportFrom( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: import-from' )

  def visit_Exec( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: exec' )

  def visit_Global( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: global' )

  def visit_Pass( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: pass' )

  def visit_Break( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: break' )

  def visit_Continue( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: continue' )

  def visit_While( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: while' )

  def visit_ExtSlice( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: extslice' )



"""
=========================================================================
test_cases.py
=========================================================================
Centralized test case repository for the SystemVerilog backend.

Author : Peitian Pan
Date   : Dec 16, 2019
"""

from textwrap import dedent

import pymtl3.passes.rtlir.rtype.RTLIRDataType as rdt
from pymtl3 import *
from pymtl3.passes.testcases import (
    Bits32Foo,
    Bits32x5Foo,
    CaseArrayBits32IfcInUpblkComp,
    CaseBits32ArrayConnectSubCompAttrComp,
    CaseBits32ArraySubCompAttrUpblkComp,
    CaseBits32BitSelUpblkComp,
    CaseBits32ConnectSubCompAttrComp,
    CaseBits32FooInBits32OutComp,
    CaseBits32SubCompAttrUpblkComp,
    CaseBits32x2ConcatComp,
    CaseBits32x2ConcatConstComp,
    CaseBits32x2ConcatFreeVarComp,
    CaseBits32x2ConcatMixedComp,
    CaseBits32x2ConcatUnpackedSignalComp,
    CaseBits64PartSelUpblkComp,
    CaseBits64SextInComp,
    CaseBits64ZextInComp,
    CaseConnectArrayBits32FooIfcComp,
    CaseConnectArrayNestedIfcComp,
    CaseConnectArrayStructAttrToOutComp,
    CaseConnectArraySubCompArrayStructIfcComp,
    CaseConnectBitsConstToOutComp,
    CaseConnectBitSelToOutComp,
    CaseConnectConstStructAttrToOutComp,
    CaseConnectConstToOutComp,
    CaseConnectInToWireComp,
    CaseConnectLiteralStructComp,
    CaseConnectNestedStructPackedArrayComp,
    CaseConnectPassThroughLongNameComp,
    CaseConnectSliceToOutComp,
    CaseConnectValRdyIfcComp,
    CaseConnectValRdyIfcUpblkComp,
    CaseConstStructInstComp,
    CaseElifBranchComp,
    CaseFixedSizeSliceComp,
    CaseForRangeLowerUpperStepPassThroughComp,
    CaseIfBasicComp,
    CaseIfBoolOpInForStmtComp,
    CaseIfDanglingElseInnerComp,
    CaseIfDanglingElseOutterComp,
    CaseIfExpInForStmtComp,
    CaseIfExpUnaryOpInForStmtComp,
    CaseIfTmpVarInForStmtComp,
    CaseInterfaceArrayNonStaticIndexComp,
    CaseLambdaConnectComp,
    CaseLambdaConnectWithListComp,
    CaseNestedIfComp,
    CaseNestedStructPackedArrayUpblkComp,
    CasePassThroughComp,
    CaseReducesInx3OutComp,
    CaseSequentialPassThroughComp,
    CaseSizeCastPaddingStructPort,
    CaseStructPackedArrayUpblkComp,
    CaseSVerilogReservedComp,
    NestedStructPackedPlusScalar,
    ThisIsABitStructWithSuperLongName,
    set_attributes,
)

CaseSizeCastPaddingStructPort = set_attributes( CaseSizeCastPaddingStructPort,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = 64'( in_ );
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo;

        module DUT
        (
          input logic [0:0] clk,
          input Bits32Foo in_,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = 64'( in_ );
          end

        endmodule
    '''
)

CasePassThroughComp = set_attributes( CasePassThroughComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_;
          end

        endmodule
    '''
)

CaseSequentialPassThroughComp = set_attributes( CaseSequentialPassThroughComp,
    'REF_UPBLK',
    '''\
        always_ff @(posedge clk) begin : upblk
          out <= in_;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_ff @(posedge clk) begin : upblk
            out <= in_;
          end

        endmodule
    '''
)

CaseConnectPassThroughLongNameComp = set_attributes( CaseConnectPassThroughLongNameComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out = in_;
    ''',
    'REF_SRC',
    '''\
        module DUT__a840bd1c84c05ea2
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = in_;

        endmodule
    ''',
    'REF_STRUCT',
    '''\
    ''',
)

CaseLambdaConnectComp = set_attributes( CaseLambdaConnectComp,
    'REF_UPBLK',
    '''\
        always_comb begin : _lambda__s_out
          out = in_ + 32'd42;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : _lambda__s_out
            out = in_ + 32'd42;
          end

        endmodule
    '''
)

CaseLambdaConnectWithListComp = set_attributes( CaseLambdaConnectWithListComp,
    'REF_UPBLK',
    '''\
        always_comb begin : _lambda__s_out_1_
          out[1] = in_ + 32'd42;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out [0:1],
          input logic [0:0] reset
        );

          always_comb begin : _lambda__s_out_1_
            out[1] = in_ + 32'd42;
          end

        endmodule
    '''
)

CaseBits32x2ConcatComp = set_attributes( CaseBits32x2ConcatComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_1, in_2 };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_1, in_2 };
          end

        endmodule
    '''
)

CaseBits32x2ConcatConstComp = set_attributes( CaseBits32x2ConcatConstComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { 32'd42, 32'd0 };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { 32'd42, 32'd0 };
          end

        endmodule
    '''
)

CaseBits32x2ConcatMixedComp = set_attributes( CaseBits32x2ConcatMixedComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_, 32'd0 };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_, 32'd0 };
          end

        endmodule
    '''
)

CaseBits64SextInComp = set_attributes( CaseBits64SextInComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { { 32 { in_[31] } }, in_ };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { { 32 { in_[31] } }, in_ };
          end

        endmodule
    '''
)

CaseBits64ZextInComp = set_attributes( CaseBits64ZextInComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { { 32 { 1'b0 } }, in_ };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          // @s.update
          // def upblk():
          //   s.out = zext( s.in_, 64 )

          always_comb begin : upblk
            out = { { 32 { 1'b0 } }, in_ };
          end

        endmodule
    '''
)

CaseBits32x2ConcatFreeVarComp = set_attributes( CaseBits32x2ConcatFreeVarComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_, __const__STATE_IDLE };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [63:0] out,
          input logic [0:0] reset
        );
          localparam [31:0] __const__STATE_IDLE = 32'd0;

          always_comb begin : upblk
            out = { in_, __const__STATE_IDLE };
          end

        endmodule
    '''
)

CaseBits32x2ConcatUnpackedSignalComp = set_attributes( CaseBits32x2ConcatUnpackedSignalComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_[0], in_[1] };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:1],
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_[0], in_[1] };
          end

        endmodule
    '''
)

CaseBits32BitSelUpblkComp = set_attributes( CaseBits32BitSelUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_[1];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [0:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_[1];
          end

        endmodule
    '''
)

CaseBits64PartSelUpblkComp = set_attributes( CaseBits64PartSelUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_[35:4];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [63:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_[35:4];
          end

        endmodule
    '''
)

CaseReducesInx3OutComp = set_attributes( CaseReducesInx3OutComp,
    'REF_UPBLK',
    '''\
        always_comb begin : v_reduce
          out = ( ( & in_1 ) & ( | in_2 ) ) | ( ^ in_3 );
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          input logic [31:0] in_3,
          output logic [0:0] out,
          input logic [0:0] reset
        );

          always_comb begin : v_reduce
            out = ( ( & in_1 ) & ( | in_2 ) ) | ( ^ in_3 );
          end

        endmodule
    '''
)

CaseIfBasicComp = set_attributes( CaseIfBasicComp,
    'REF_UPBLK',
    '''\
        always_comb begin : if_basic
          if ( in_[7:0] == 8'd255 ) begin
            out = in_[15:8];
          end
          else
            out = 8'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [15:0] in_,
          output logic [7:0] out,
          input logic [0:0] reset
        );

          always_comb begin : if_basic
            if ( in_[7:0] == 8'd255 ) begin
              out = in_[15:8];
            end
            else
              out = 8'd0;
          end

        endmodule
    '''
)

CaseIfDanglingElseInnerComp = set_attributes( CaseIfDanglingElseInnerComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          if ( 1'd1 ) begin
            if ( 1'd0 ) begin
              out = in_1;
            end
            else
              out = in_2;
          end
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            if ( 1'd1 ) begin
              if ( 1'd0 ) begin
                out = in_1;
              end
              else
                out = in_2;
            end
          end

        endmodule
    '''
)

CaseIfDanglingElseOutterComp = set_attributes( CaseIfDanglingElseOutterComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          if ( 1'd1 ) begin
            if ( 1'd0 ) begin
              out = in_1;
            end
          end
          else
            out = in_2;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            if ( 1'd1 ) begin
              if ( 1'd0 ) begin
                out = in_1;
              end
            end
            else
              out = in_2;
          end

        endmodule
    '''
)

CaseElifBranchComp = set_attributes( CaseElifBranchComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          if ( 1'd1 ) begin
            out = in_1;
          end
          else if ( 1'd0 ) begin
            out = in_2;
          end
          else
            out = in_3;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          input logic [31:0] in_3,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            if ( 1'd1 ) begin
              out = in_1;
            end
            else if ( 1'd0 ) begin
              out = in_2;
            end
            else
              out = in_3;
          end

        endmodule
    '''
)

CaseNestedIfComp = set_attributes( CaseNestedIfComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          if ( 1'd1 ) begin
            if ( 1'd0 ) begin
              out = in_1;
            end
            else
              out = in_2;
          end
          else if ( 1'd0 ) begin
            if ( 1'd1 ) begin
              out = in_2;
            end
            else
              out = in_3;
          end
          else if ( 1'd1 ) begin
            out = in_3;
          end
          else
            out = in_1;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          input logic [31:0] in_3,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            if ( 1'd1 ) begin
              if ( 1'd0 ) begin
                out = in_1;
              end
              else
                out = in_2;
            end
            else if ( 1'd0 ) begin
              if ( 1'd1 ) begin
                out = in_2;
              end
              else
                out = in_3;
            end
            else if ( 1'd1 ) begin
              out = in_3;
            end
            else
              out = in_1;
          end

        endmodule
    '''
)

CaseForRangeLowerUpperStepPassThroughComp = set_attributes( CaseForRangeLowerUpperStepPassThroughComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int i = 0; i < 5; i += 2 )
            out[i] = in_[i];
          for ( int i = 1; i < 5; i += 2 )
            out[i] = in_[i];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out [0:4],
          input logic [0:0] reset
        );

          always_comb begin : upblk
            for ( int i = 0; i < 5; i += 2 )
              out[i] = in_[i];
            for ( int i = 1; i < 5; i += 2 )
              out[i] = in_[i];
          end

        endmodule
    '''
)

CaseIfExpInForStmtComp = set_attributes( CaseIfExpInForStmtComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int i = 0; i < 5; i += 1 )
            out[i] = ( i == 1 ) ? in_[i] : in_[0];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out [0:4],
          input logic [0:0] reset
        );

          always_comb begin : upblk
            for ( int i = 0; i < 5; i += 1 )
              out[i] = ( i == 1 ) ? in_[i] : in_[0];
          end

        endmodule
    '''
)

CaseIfExpUnaryOpInForStmtComp = set_attributes( CaseIfExpUnaryOpInForStmtComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int i = 0; i < 5; i += 1 )
            out[i] = ( i == 1 ) ? ~in_[i] : in_[0];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out [0:4],
          input logic [0:0] reset
        );

          always_comb begin : upblk
            for ( int i = 0; i < 5; i += 1 )
              out[i] = ( i == 1 ) ? ~in_[i] : in_[0];
          end

        endmodule
    '''
)

CaseIfBoolOpInForStmtComp = set_attributes( CaseIfBoolOpInForStmtComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int i = 0; i < 5; i += 1 )
            if ( ( in_[i] != 32'd0 ) && ( ( i < 4 ) ? in_[i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
              out[i] = in_[i];
            end
            else
              out[i] = 32'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out [0:4],
          input logic [0:0] reset
        );

          always_comb begin : upblk
            for ( int i = 0; i < 5; i += 1 )
              if ( ( in_[i] != 32'd0 ) && ( ( i < 4 ) ? in_[i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
                out[i] = in_[i];
              end
              else
                out[i] = 32'd0;
          end

        endmodule
    '''
)

CaseIfTmpVarInForStmtComp = set_attributes( CaseIfTmpVarInForStmtComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int i = 0; i < 5; i += 1 ) begin
            if ( ( in_[i] != 32'd0 ) && ( ( i < 4 ) ? in_[i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
              __tmpvar__upblk_tmpvar = in_[i];
            end
            else
              __tmpvar__upblk_tmpvar = 32'd0;
            out[i] = __tmpvar__upblk_tmpvar;
          end
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out [0:4],
          input logic [0:0] reset
        );
          logic [31:0] __tmpvar__upblk_tmpvar;

          always_comb begin : upblk
            for ( int i = 0; i < 5; i += 1 ) begin
              if ( ( in_[i] != 32'd0 ) && ( ( i < 4 ) ? in_[i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
                __tmpvar__upblk_tmpvar = in_[i];
              end
              else
                __tmpvar__upblk_tmpvar = 32'd0;
              out[i] = __tmpvar__upblk_tmpvar;
            end
          end

        endmodule
    '''
)

CaseFixedSizeSliceComp = set_attributes( CaseFixedSizeSliceComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int i = 0; i < 2; i += 1 )
            out[i] = in_[i * 8 +: 8];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [15:0] in_,
          output logic [7:0] out [0:1],
          input logic [0:0] reset
        );

          always_comb begin : upblk
            for ( int i = 0; i < 2; i += 1 )
              out[i] = in_[i * 8 +: 8];
          end

        endmodule
    '''
)

CaseBits32FooInBits32OutComp = set_attributes( CaseBits32FooInBits32OutComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_.foo;
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo;

        module DUT
        (
          input logic [0:0] clk,
          input Bits32Foo in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_.foo;
          end

        endmodule
    '''
)

CaseConstStructInstComp = set_attributes( CaseConstStructInstComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_.foo;
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo;

        module DUT
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          localparam Bits32Foo in_ = { 32'd0 };

          always_comb begin : upblk
            out = in_.foo;
          end

        endmodule
    '''
)

CaseStructPackedArrayUpblkComp = set_attributes( CaseStructPackedArrayUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_.foo[0], in_.foo[1], in_.foo[2] };
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [4:0][31:0] foo;
        } Bits32x5Foo;

        module DUT
        (
          input logic [0:0] clk,
          input Bits32x5Foo in_,
          output logic [95:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_.foo[0], in_.foo[1], in_.foo[2] };
          end

        endmodule
    '''
)

CaseNestedStructPackedArrayUpblkComp = set_attributes( CaseNestedStructPackedArrayUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_.bar[0], in_.woo.foo, in_.foo };
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo;

        typedef struct packed {
          logic [31:0] foo;
          logic [1:0][31:0] bar;
          Bits32Foo woo;
        } NestedStructPackedPlusScalar;

        module DUT
        (
          input logic [0:0] clk,
          input NestedStructPackedPlusScalar in_,
          output logic [95:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_.bar[0], in_.woo.foo, in_.foo };
          end

        endmodule
    '''
)

CaseConnectValRdyIfcUpblkComp = set_attributes( CaseConnectValRdyIfcUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out__val = in___val;
          out__msg = in___msg;
          in___rdy = out__rdy;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [0:0] reset,
          input logic [31:0] in___msg,
          output logic [0:0] in___rdy,
          input logic [0:0] in___val,
          output logic [31:0] out__msg,
          input logic [0:0] out__rdy,
          output logic [0:0] out__val
        );

          always_comb begin : upblk
            out__val = in___val;
            out__msg = in___msg;
            in___rdy = out__rdy;
          end

        endmodule
    '''
)

CaseArrayBits32IfcInUpblkComp = set_attributes( CaseArrayBits32IfcInUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in___1__foo;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset,
          input logic [31:0] in___0__foo,
          input logic [31:0] in___1__foo,
          input logic [31:0] in___2__foo,
          input logic [31:0] in___3__foo,
          input logic [31:0] in___4__foo
        );

          always_comb begin : upblk
            out = in___1__foo;
          end

        endmodule
    '''
)

CaseBits32SubCompAttrUpblkComp = set_attributes( CaseBits32SubCompAttrUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = b__out;
        end
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule

        module DUT
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [0:0] b__clk;
          logic [31:0] b__out;
          logic [0:0] b__reset;

          Bits32OutDrivenComp b
          (
            .clk( b__clk ),
            .out( b__out ),
            .reset( b__reset )
          );

          always_comb begin : upblk
            out = b__out;
          end

          assign b__clk = clk;
          assign b__reset = reset;

        endmodule
    '''
)

CaseBits32ArraySubCompAttrUpblkComp = set_attributes( CaseBits32ArraySubCompAttrUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = b__1__out;
        end
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule

        module DUT
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [0:0] b__0__clk;
          logic [31:0] b__0__out;
          logic [0:0] b__0__reset;

          Bits32OutDrivenComp b__0
          (
            .clk( b__0__clk ),
            .out( b__0__out ),
            .reset( b__0__reset )
          );

          logic [0:0] b__1__clk;
          logic [31:0] b__1__out;
          logic [0:0] b__1__reset;

          Bits32OutDrivenComp b__1
          (
            .clk( b__1__clk ),
            .out( b__1__out ),
            .reset( b__1__reset )
          );

          logic [0:0] b__2__clk;
          logic [31:0] b__2__out;
          logic [0:0] b__2__reset;

          Bits32OutDrivenComp b__2
          (
            .clk( b__2__clk ),
            .out( b__2__out ),
            .reset( b__2__reset )
          );

          logic [0:0] b__3__clk;
          logic [31:0] b__3__out;
          logic [0:0] b__3__reset;

          Bits32OutDrivenComp b__3
          (
            .clk( b__3__clk ),
            .out( b__3__out ),
            .reset( b__3__reset )
          );

          logic [0:0] b__4__clk;
          logic [31:0] b__4__out;
          logic [0:0] b__4__reset;

          Bits32OutDrivenComp b__4
          (
            .clk( b__4__clk ),
            .out( b__4__out ),
            .reset( b__4__reset )
          );

          always_comb begin : upblk
            out = b__1__out;
          end

          assign b__0__clk = clk;
          assign b__0__reset = reset;
          assign b__1__clk = clk;
          assign b__1__reset = reset;
          assign b__2__clk = clk;
          assign b__2__reset = reset;
          assign b__3__clk = clk;
          assign b__3__reset = reset;
          assign b__4__clk = clk;
          assign b__4__reset = reset;

        endmodule
    '''
)

CaseConnectInToWireComp = set_attributes( CaseConnectInToWireComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_ [0:4],
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '''\
        logic [31:0] wire_ [0:4];
    ''',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = wire_[2];
        assign wire_[0] = in_[0];
        assign wire_[1] = in_[1];
        assign wire_[2] = in_[2];
        assign wire_[3] = in_[3];
        assign wire_[4] = in_[4];
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [31:0] wire_ [0:4];

          assign out = wire_[2];
          assign wire_[0] = in_[0];
          assign wire_[1] = in_[1];
          assign wire_[2] = in_[2];
          assign wire_[3] = in_[3];
          assign wire_[4] = in_[4];

        endmodule
    '''
)

CaseConnectBitsConstToOutComp = set_attributes( CaseConnectBitsConstToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = 32'd0;
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd0;

        endmodule
    '''
)

CaseConnectConstToOutComp = set_attributes( CaseConnectConstToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '''\
        localparam [31:0] const_ [0:4] = '{ 32'd42, 32'd42, 32'd42, 32'd42, 32'd42 };
    ''',
    'REF_CONN',
    '''\
        assign out = 32'd42;
    ''',
    'REF_SRC',
    # const_ is not used in upblks so will be optimized away
    '''\
        module DUT
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule
    '''
)

CaseConnectBitSelToOutComp = set_attributes( CaseConnectBitSelToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [0:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = in_[0:0];
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [0:0] out,
          input logic [0:0] reset
        );

          assign out = in_[0:0];

        endmodule
    '''
)

CaseConnectSliceToOutComp = set_attributes( CaseConnectSliceToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [3:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = in_[7:4];
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [3:0] out,
          input logic [0:0] reset
        );

          assign out = in_[7:4];

        endmodule
    '''
)

CaseConnectConstStructAttrToOutComp = set_attributes( CaseConnectConstStructAttrToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONN',
    '''\
        assign out = 32'd42;
    ''',
    'REF_STRUCT',
    (
        rdt.Struct('Bits32Foo', {'foo':rdt.Vector(32)}),
        dedent('''\
                  typedef struct packed {
                    logic [31:0] foo;
                  } Bits32Foo;
               ''')
    ),
    'REF_SRC',
    # struct definition will be eliminated because it's not used
    # in an upblk
    '''\
        module DUT
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule
    '''
)

CaseConnectLiteralStructComp = set_attributes( CaseConnectLiteralStructComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        output NestedStructPackedPlusScalar out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONN',
    '''\
        assign out = { 32'd42, { { 32'd2, 32'd1 } }, { 32'd3 } };
    ''',
    'REF_STRUCT',
    (
        rdt.Struct('NestedStructPackedPlusScalar', {
          'foo':rdt.Vector(32),
          'bar':rdt.PackedArray([2], rdt.Vector(32)),
          'woo':rdt.Struct('Bits32Foo', {'foo':rdt.Vector(32)}),
        }),
        dedent('''\
                  typedef struct packed {
                    logic [31:0] foo;
                    logic [1:0][31:0] bar;
                    Bits32Foo woo;
                  } NestedStructPackedPlusScalar;
               ''')
    ),
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo;

        typedef struct packed {
          logic [31:0] foo;
          logic [1:0][31:0] bar;
          Bits32Foo woo;
        } NestedStructPackedPlusScalar;

        module DUT
        (
          input logic [0:0] clk,
          output NestedStructPackedPlusScalar out,
          input logic [0:0] reset
        );

          assign out = { 32'd42, { { 32'd2, 32'd1 } }, { 32'd3 } };

        endmodule
    '''
)

CaseConnectArrayStructAttrToOutComp = set_attributes( CaseConnectArrayStructAttrToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input Bits32x5Foo in_,
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONN',
    '''\
        assign out = in_.foo[1];
    ''',
    'REF_STRUCT',
    (
        rdt.Struct('Bits32x5Foo', {'foo':rdt.PackedArray([5], rdt.Vector(32))}),
        dedent('''\
                  typedef struct packed {
                    logic [4:0][31:0] foo;
                  } Bits32x5Foo;
               ''')
    ),
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [4:0][31:0] foo;
        } Bits32x5Foo;

        module DUT
        (
          input logic [0:0] clk,
          input Bits32x5Foo in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = in_.foo[1];

        endmodule
    '''
)

CaseConnectNestedStructPackedArrayComp = set_attributes( CaseConnectNestedStructPackedArrayComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input NestedStructPackedPlusScalar in_,
        output logic [95:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONN',
    '''\
        assign out[31:0] = in_.foo;
        assign out[63:32] = in_.woo.foo;
        assign out[95:64] = in_.bar[0];
    ''',
    'REF_STRUCT',
    (
        rdt.Struct('NestedStructPackedPlusScalar', {
          'foo':rdt.Vector(32),
          'bar':rdt.PackedArray([2], rdt.Vector(32)),
          'woo':rdt.Struct('Bits32Foo', {'foo':rdt.Vector(32)}),
        }),
        dedent('''\
                  typedef struct packed {
                    logic [31:0] foo;
                    logic [1:0][31:0] bar;
                    Bits32Foo woo;
                  } NestedStructPackedPlusScalar;
               ''')
    ),
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo;

        typedef struct packed {
          logic [31:0] foo;
          logic [1:0][31:0] bar;
          Bits32Foo woo;
        } NestedStructPackedPlusScalar;

        module DUT
        (
          input logic [0:0] clk,
          input NestedStructPackedPlusScalar in_,
          output logic [95:0] out,
          input logic [0:0] reset
        );

          assign out[31:0] = in_.foo;
          assign out[63:32] = in_.woo.foo;
          assign out[95:64] = in_.bar[0];

        endmodule
    '''
)

CaseConnectValRdyIfcComp = set_attributes( CaseConnectValRdyIfcComp,
    'REF_IFC',
    '''\
        input  logic [31:0] in___msg,
        output logic [0:0]  in___rdy,
        input  logic [0:0]  in___val,
        output logic [31:0] out__msg,
        input  logic [0:0]  out__rdy,
        output logic [0:0]  out__val
    ''',
    'REF_CONN',
    '''\
        assign out__msg = in___msg;
        assign in___rdy = out__rdy;
        assign out__val = in___val;
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [0:0] reset,
          input logic [31:0] in___msg,
          output logic [0:0] in___rdy,
          input logic [0:0] in___val,
          output logic [31:0] out__msg,
          input logic [0:0] out__rdy,
          output logic [0:0] out__val
        );

          assign out__msg = in___msg;
          assign in___rdy = out__rdy;
          assign out__val = in___val;

        endmodule
    '''
)

CaseConnectArrayBits32FooIfcComp = set_attributes( CaseConnectArrayBits32FooIfcComp,
    'REF_IFC',
    '''\
        input Bits32Foo  in___0__foo,
        input Bits32Foo  in___1__foo,
        output Bits32Foo  out__0__foo,
        output Bits32Foo  out__1__foo
    ''',
    'REF_CONN',
    '''\
        assign out__0__foo = in___0__foo;
        assign out__1__foo = in___1__foo;
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo;

        module DUT
        (
          input logic [0:0] clk,
          input logic [0:0] reset,
          input Bits32Foo in___0__foo,
          input Bits32Foo in___1__foo,
          output Bits32Foo out__0__foo,
          output Bits32Foo out__1__foo
        );

          assign out__0__foo = in___0__foo;
          assign out__1__foo = in___1__foo;

        endmodule
    '''
)

CaseConnectArrayNestedIfcComp = set_attributes( CaseConnectArrayNestedIfcComp,
    'REF_IFC',
    '''\
        input  logic [0:0]  in___0__ctrl_foo,
        input  logic [31:0] in___0__memifc__msg,
        output logic [0:0]  in___0__memifc__rdy,
        input  logic [0:0]  in___0__memifc__val,
        input  logic [0:0]  in___1__ctrl_foo,
        input  logic [31:0] in___1__memifc__msg,
        output logic [0:0]  in___1__memifc__rdy,
        input  logic [0:0]  in___1__memifc__val,

        output  logic [0:0]  out__0__ctrl_foo,
        output  logic [31:0] out__0__memifc__msg,
        input   logic [0:0]  out__0__memifc__rdy,
        output  logic [0:0]  out__0__memifc__val,
        output  logic [0:0]  out__1__ctrl_foo,
        output  logic [31:0] out__1__memifc__msg,
        input   logic [0:0]  out__1__memifc__rdy,
        output  logic [0:0]  out__1__memifc__val
    ''',
    'REF_CONN',
    '''\
        assign out__0__ctrl_foo = in___0__ctrl_foo;
        assign out__0__memifc__msg = in___0__memifc__msg;
        assign in___0__memifc__rdy = out__0__memifc__rdy;
        assign out__0__memifc__val = in___0__memifc__val;
        assign out__1__ctrl_foo = in___1__ctrl_foo;
        assign out__1__memifc__msg = in___1__memifc__msg;
        assign in___1__memifc__rdy = out__1__memifc__rdy;
        assign out__1__memifc__val = in___1__memifc__val;
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [0:0] reset,
          input logic [0:0] in___0__ctrl_foo,
          input logic [31:0] in___0__memifc__msg,
          output logic [0:0] in___0__memifc__rdy,
          input logic [0:0] in___0__memifc__val,
          input logic [0:0] in___1__ctrl_foo,
          input logic [31:0] in___1__memifc__msg,
          output logic [0:0] in___1__memifc__rdy,
          input logic [0:0] in___1__memifc__val,
          output logic [0:0] out__0__ctrl_foo,
          output logic [31:0] out__0__memifc__msg,
          input logic [0:0] out__0__memifc__rdy,
          output logic [0:0] out__0__memifc__val,
          output logic [0:0] out__1__ctrl_foo,
          output logic [31:0] out__1__memifc__msg,
          input logic [0:0] out__1__memifc__rdy,
          output logic [0:0] out__1__memifc__val
        );

          assign out__0__ctrl_foo = in___0__ctrl_foo;
          assign out__0__memifc__msg = in___0__memifc__msg;
          assign in___0__memifc__rdy = out__0__memifc__rdy;
          assign out__0__memifc__val = in___0__memifc__val;
          assign out__1__ctrl_foo = in___1__ctrl_foo;
          assign out__1__memifc__msg = in___1__memifc__msg;
          assign in___1__memifc__rdy = out__1__memifc__rdy;
          assign out__1__memifc__val = in___1__memifc__val;

        endmodule
    '''
)

CaseBits32ConnectSubCompAttrComp = set_attributes( CaseBits32ConnectSubCompAttrComp,
    'REF_COMP',
    '''\
        logic [0:0] b__clk;
        logic [31:0] b__out;
        logic [0:0] b__reset;

        Bits32OutDrivenComp b
        (
          .clk( b__clk ),
          .out( b__out ),
          .reset( b__reset )
        );
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule

        module DUT
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [0:0] b__clk;
          logic [31:0] b__out;
          logic [0:0] b__reset;

          Bits32OutDrivenComp b
          (
            .clk( b__clk ),
            .out( b__out ),
            .reset( b__reset )
          );

          assign b__clk = clk;
          assign b__reset = reset;
          assign out = b__out;

        endmodule
    '''
)

CaseConnectArraySubCompArrayStructIfcComp = set_attributes( CaseConnectArraySubCompArrayStructIfcComp,
    'REF_COMP',
    '''\
        logic [0:0] b__0__clk;
        logic [31:0] b__0__out;
        logic [0:0] b__0__reset;
        Bits32Foo b__0__ifc__0__foo [0:0];

        Bits32ArrayStructIfcComp b__0
        (
          .clk( b__0__clk ),
          .out( b__0__out ),
          .reset( b__0__reset ),
          .ifc__0__foo( b__0__ifc__0__foo )
        );
    ''',
    'REF_SRC',
    '''\
      typedef struct packed {
        logic [31:0] foo;
      } Bits32Foo;

      module Bits32ArrayStructIfcComp
      (
        input logic [0:0] clk,
        output logic [31:0] out,
        input logic [0:0] reset,
        input Bits32Foo ifc__0__foo [0:0]
      );

        assign out = ifc__0__foo[0].foo;

      endmodule

      module DUT
      (
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [31:0] out,
        input logic [0:0] reset
      );
        logic [0:0] b__0__clk;
        logic [31:0] b__0__out;
        logic [0:0] b__0__reset;
        Bits32Foo b__0__ifc__0__foo [0:0];

        Bits32ArrayStructIfcComp b__0
        (
          .clk( b__0__clk ),
          .out( b__0__out ),
          .reset( b__0__reset ),
          .ifc__0__foo( b__0__ifc__0__foo )
        );

        assign b__0__clk = clk;
        assign b__0__reset = reset;
        assign b__0__ifc__0__foo[0].foo = in_;
        assign out = b__0__out;

      endmodule
    '''
)

CaseBits32ArrayConnectSubCompAttrComp = set_attributes( CaseBits32ArrayConnectSubCompAttrComp,
    'REF_COMP',
    '''\
        logic [0:0] b__0__clk;
        logic [31:0] b__0__out;
        logic [0:0] b__0__reset;

        Bits32OutDrivenComp b__0
        (
          .clk( b__0__clk ),
          .out( b__0__out ),
          .reset( b__0__reset )
        );

        logic [0:0] b__1__clk;
        logic [31:0] b__1__out;
        logic [0:0] b__1__reset;

        Bits32OutDrivenComp b__1
        (
          .clk( b__1__clk ),
          .out( b__1__out ),
          .reset( b__1__reset )
        );

        logic [0:0] b__2__clk;
        logic [31:0] b__2__out;
        logic [0:0] b__2__reset;

        Bits32OutDrivenComp b__2
        (
          .clk( b__2__clk ),
          .out( b__2__out ),
          .reset( b__2__reset )
        );

        logic [0:0] b__3__clk;
        logic [31:0] b__3__out;
        logic [0:0] b__3__reset;

        Bits32OutDrivenComp b__3
        (
          .clk( b__3__clk ),
          .out( b__3__out ),
          .reset( b__3__reset )
        );

        logic [0:0] b__4__clk;
        logic [31:0] b__4__out;
        logic [0:0] b__4__reset;

        Bits32OutDrivenComp b__4
        (
          .clk( b__4__clk ),
          .out( b__4__out ),
          .reset( b__4__reset )
        );
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule

        module DUT
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [0:0] b__0__clk;
          logic [31:0] b__0__out;
          logic [0:0] b__0__reset;

          Bits32OutDrivenComp b__0
          (
            .clk( b__0__clk ),
            .out( b__0__out ),
            .reset( b__0__reset )
          );

          logic [0:0] b__1__clk;
          logic [31:0] b__1__out;
          logic [0:0] b__1__reset;

          Bits32OutDrivenComp b__1
          (
            .clk( b__1__clk ),
            .out( b__1__out ),
            .reset( b__1__reset )
          );

          logic [0:0] b__2__clk;
          logic [31:0] b__2__out;
          logic [0:0] b__2__reset;

          Bits32OutDrivenComp b__2
          (
            .clk( b__2__clk ),
            .out( b__2__out ),
            .reset( b__2__reset )
          );

          logic [0:0] b__3__clk;
          logic [31:0] b__3__out;
          logic [0:0] b__3__reset;

          Bits32OutDrivenComp b__3
          (
            .clk( b__3__clk ),
            .out( b__3__out ),
            .reset( b__3__reset )
          );

          logic [0:0] b__4__clk;
          logic [31:0] b__4__out;
          logic [0:0] b__4__reset;

          Bits32OutDrivenComp b__4
          (
            .clk( b__4__clk ),
            .out( b__4__out ),
            .reset( b__4__reset )
          );

          assign b__0__clk = clk;
          assign b__0__reset = reset;
          assign b__1__clk = clk;
          assign b__1__reset = reset;
          assign b__2__clk = clk;
          assign b__2__reset = reset;
          assign b__3__clk = clk;
          assign b__3__reset = reset;
          assign b__4__clk = clk;
          assign b__4__reset = reset;
          assign out = b__1__out;

        endmodule
    '''
)



#=========================================================================
# TranslationPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 12, 2019
"""Translate a PyMTL component hierarhcy into SystemVerilog source code."""

import filecmp
import os

from pymtl3.passes.BasePass import BasePass, PassMetadata

from .SVTranslator import SVTranslator


def mk_TranslationPass( _SVTranslator ):

  class _TranslationPass( BasePass ):

    def __call__( s, top ):

      s.top = top
      s.translator = _SVTranslator( s.top )
      s.traverse_hierarchy( top )

    def traverse_hierarchy( s, m ):

      if hasattr(m, "sverilog_translate") and m.sverilog_translate:

        if not hasattr( m, '_pass_sverilog_translation' ):
          m._pass_sverilog_translation = PassMetadata()

        s.translator.translate( m )

        module_name = s.translator._top_module_full_name
        output_file = module_name + '.sv'
        temporary_file = module_name + '.sv.tmp'

        # First write the file to a temporary file
        m._pass_sverilog_translation.is_same = False
        with open( temporary_file, 'w' ) as output:
          output.write( s.translator.hierarchy.src )
          output.flush()
          os.fsync( output )
          output.close()

        # `is_same` is set if there exists a file that has the same filename as
        # `output_file`, and that file is the same as the temporary file
        if ( os.path.exists(output_file) ):
          m._pass_sverilog_translation.is_same = \
              filecmp.cmp( temporary_file, output_file )

        # Rename the temporary file to the output file
        os.rename( temporary_file, output_file )

        # Expose some attributes about the translation process
        m.translated_top_module_name = module_name
        m._translator = s.translator
        m._pass_sverilog_translation.translated = True

      else:
        for child in m.get_child_components():
          s.traverse_hierarchy( child )

  return _TranslationPass

TranslationPass = mk_TranslationPass( SVTranslator )



from pymtl3.datatypes import Bits32
from pymtl3.dsl import *

from ..PassGroups import UnrollSim


def test_very_deep_dag():

  class Inner(Component):
    def construct( s ):
      s.in_ = InPort(Bits32)
      s.out = OutPort(Bits32)

      @s.update
      def up():
        s.out = s.in_ + 1

    def done( s ):
      return True

    def line_trace( s ):
      return "{} > {}".format( s.a, s.b, s.c, s.d )

  class Top(Component):
    def construct( s, N=2000 ):
      s.inners = [ Inner() for i in range(N) ]
      for i in range(N-1):
        s.inners[i].out //= s.inners[i+1].in_

      s.out = OutPort(Bits32)
      @s.update_ff
      def ff():
        if s.reset:
          s.out <<= 0
        else:
          s.out <<= s.out + s.inners[N-1].out

    def line_trace( s ):
      return str(s.inners[-1].out) + " " + str(s.out)

  N = 2000
  A = Top( N )

  A.apply( UnrollSim() )

  T = 0
  while T < 5:
    A.tick()
    print(A.line_trace())
    assert A.out == T * N
    T += 1
  return A



#========================================================================
# LineTraceParamPass.py
#========================================================================
# Enable multi-level line trace.
#
# Author : Yanghui Ou
#   Date : May 29, 2019

from pymtl3.dsl import *
from pymtl3.passes.BasePass import BasePass, PassMetadata


class LineTraceParamPass( BasePass ):

  def __call__( self, top ):

    def wrap_line_trace( obj ):
      if not hasattr( obj, '_ml_trace' ):
        obj._ml_trace = PassMetadata()
      obj._ml_trace.line_trace = obj.line_trace

      def wrapped_line_trace( self, *args, **kwargs ):
        if self._dsl.param_tree is not None:
          if self._dsl.param_tree.leaf is not None:
            if 'line_trace' in self._dsl.param_tree.leaf:
              # TODO: figure out whether it is necessary to enforce no
              # positional args.
              assert len( args ) == 0
              more_args = self._dsl.param_tree.leaf['line_trace'].items()
              kwargs.update({ x:y for x, y in more_args })
        return self._ml_trace.line_trace( *args, **kwargs )

      obj.line_trace = lambda *args, **kwargs : wrapped_line_trace( obj, *args, **kwargs )

    all_objects = top.get_all_object_filter( lambda x: True )
    for obj in all_objects:
      if hasattr( obj, 'line_trace' ):
        wrap_line_trace( obj )



#=========================================================================
# TranslationImport_dynlib_close_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Dec 6, 2019
"""Test if the shared library was closed at the end of simulation."""

from pymtl3 import *
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test import TestVectorSimulator

from .. import TranslationImportPass
from ..import_.ImportConfigs import ImportConfigs
from ..util.utility import get_file_hash


def run_test( _m ):
  try:
    _m.elaborate()
    _m.sverilog_translate_import = True
    _m.config_sverilog_import = \
        ImportConfigs(vl_Wno_list=['UNOPTFLAT', 'UNSIGNED'], vl_trace=True)
    m = TranslationImportPass()( _m )
    sim = TestVectorSimulator( m, _m._test_vectors, _m._tv_in, _m._tv_out )
    sim.run_test()
  finally:
    try:
      m.finalize()
    except UnboundLocalError:
      # This test fails due to translation errors
      pass

def test_dynlib_close():
  class Comb:
    class A( Component ):
      def construct( s ):
        s.in_ = InPort( Bits32 )
        s.out = OutPort( Bits32 )
        @s.update
        def upblk():
          s.out = s.in_
  class Seq:
    class A( Component ):
      def construct( s ):
        s.in_ = InPort( Bits32 )
        s.out = OutPort( Bits32 )
        @s.update_ff
        def upblk():
          s.out <<= s.in_

  comb_a = Comb.A()
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[1])
  comb_a._test_vectors = [
    [    0,   0  ],
    [   42,   42 ],
    [   24,   24 ],
    [   -2,   -2 ],
    [   -1,   -1 ],
  ]
  comb_a._tv_in, comb_a._tv_out = tv_in, tv_out
  run_test( comb_a )
  comb_hash = get_file_hash('A.verilator1.vcd')

  seq_a = Seq.A()
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    if tv[1] != '*':
      assert m.out == Bits32(tv[1])
  seq_a._test_vectors = [
    [    0,   '*' ],
    [   42,    0  ],
    [   24,    42 ],
    [   -2,    24 ],
    [   -1,    -2 ],
  ]
  seq_a._tv_in, seq_a._tv_out = tv_in, tv_out
  run_test( seq_a )
  seq_hash = get_file_hash('A.verilator1.vcd')

  assert comb_hash != seq_hash



"""
========================================================================
XcelMsg.py
========================================================================
Accelerator message type implementation.

Author : Yanghui Ou
Date   : June 3, 2019
"""
from pymtl3 import *


def mk_xcel_msg( addr, data ):
  return mk_xcel_req_msg( addr, data ), mk_xcel_resp_msg( data )

def mk_xcel_req_msg( addr, data ):
  AddrType = mk_bits( addr )
  DataType = mk_bits( data )
  cls_name = "XcelReqMsg_{}_{}".format( addr, data )

  def req_to_str( self ):
    return "{}:{}:{}".format(
      "rd" if self.type_ == XcelMsgType.READ else "wr",
      AddrType( self.addr ),
      DataType( self.data ) if self.type_ != XcelMsgType.READ else
      " " * ( data//4 ),
    )

  req_cls = mk_bitstruct( cls_name, {
    'type_': Bits1,
    'addr':  AddrType,
    'data':  DataType,
  },
  namespace = {
    '__str__' : req_to_str
  })
  return req_cls

def mk_xcel_resp_msg( data ):
  DataType = mk_bits( data )
  cls_name = "XcelRespMsg_{}".format( data )

  def resp_to_str( self ):
    return "{}:{}".format(
      "rd" if self.type_ == XcelMsgType.READ else "wr",
      DataType( self.data ) if self.type_ != XcelMsgType.WRITE else
      " " * ( data//4 ),
    )

  resp_cls = mk_bitstruct( cls_name, {
    'type_': Bits1,
    'data':  DataType,
  },
  namespace = {
    '__str__' : resp_to_str
  })
  return resp_cls

class XcelMsgType:
  # TODO: figure out whether we want to use Bits1 here.
  READ  = Bits1(0)
  WRITE = Bits1(1)



"""
========================================================================
errors.py
========================================================================
Errors thrown by other files in this folder.

Author : Shunning Jiang
Date   : Dec 27, 2018
"""
import inspect


class PyMTLDeprecationError( Exception ):
  """ Raise when a deprecated API is called"""

class SignalTypeError( Exception ):
  """ Raise when a declared signal is of wrong type """

class MultiWriterError( Exception ):
  """ Raise when a variable is written by multiple update blocks/nets """

class NoWriterError( Exception ):
  """ Raise when a net has no writer (driver) """
  def __init__( self, nets ):
    return super().__init__( \
    "The following nets need drivers.\nNet:\n - {} ".format(
      "\nNet:\n - ".join( [ "\n - ".join( [ repr(x) for x in y ] )
                            for y in nets ]) ) )

class InvalidUpblkWriteError( Exception ):
  """ In update/update_ff, raise when a component is assigned """
  def __init__( self, hostobj, blk, lineno, obj ):

    filepath = inspect.getfile( hostobj.__class__ )
    blk_src, base_lineno  = inspect.getsourcelines( blk )

    # Shunning: we need to subtract 1 from inspect's lineno when we add it
    # to base_lineno because it starts from 1!
    lineno -= 1
    error_lineno = base_lineno + lineno

    return super().__init__( \
"""
In file {}:{} in {}

{} {}
^^^ Cannot Assign to {} of type {}.
In update block, assigning values to components/interfaces/method ports is forbidden.
(when constructing instance {} of class \"{}\" in the hierarchy)

Suggestion: check the declaration of the variables, or fix this assignment.""".format( \
      filepath, error_lineno, blk.__name__,
      error_lineno, blk_src[ lineno ].lstrip(''),
      repr(obj), obj.__class__.__name__,
      repr(hostobj), hostobj.__class__.__name__ )
    )

class InvalidFFAssignError( Exception ):
  """ In update_ff, raise when signal is not <<= -ed, or temp is not = -ed """
  def __init__( self, hostobj, blk, lineno, suggestion ):

    filepath = inspect.getfile( hostobj.__class__ )
    blk_src, base_lineno  = inspect.getsourcelines( blk )

    # Shunning: we need to subtract 1 from inspect's lineno when we add it
    # to base_lineno because it starts from 1!
    lineno -= 1
    error_lineno = base_lineno + lineno

    return super().__init__( \
"""
In file {}:{} in {}

{} {}
^^^ In update_ff, we only allow <<= to valid fields for constructing nonblocking assignments.
(when constructing instance {} of class \"{}\" in the hierarchy)

Suggestion: Line {} {}""".format( \
      filepath, error_lineno, blk.__name__,
      error_lineno, blk_src[ lineno ].lstrip(''),
      repr(hostobj), hostobj.__class__.__name__,
      error_lineno, suggestion )
    )

class VarNotDeclaredError( Exception ):
  """ Raise when a variable in an update block is not declared """
  def __init__( self, obj, field, blk=None, blk_hostobj=None, lineno=0 ):

    if not blk:
      return super().__init__() # this is just temporary message

    filepath = inspect.getfile( blk_hostobj.__class__ )
    blk_src, base_lineno  = inspect.getsourcelines( blk )

    # Shunning: we need to subtract 1 from inspect's lineno when we add it
    # to base_lineno because it starts from 1!
    lineno -= 1
    error_lineno = base_lineno + lineno

    return super().__init__( \
"""
In file {}:{} in {}

{} {}
^^^ Field \"{}\" of object \"{}\" (class \"{}\") is accessed in block \"{}\",
    but {} does not have field \"{}\".
(when constructing instance {} of class \"{}\" in the hierarchy)

Suggestion: fix incorrect field access at line {}, or fix the declaration somewhere.""".format( \
      filepath, error_lineno, blk.__name__,
      error_lineno, blk_src[ lineno ].lstrip(''),
      field, repr(obj), obj.__class__.__name__, blk.__name__,
      repr(obj), field,
      repr(blk_hostobj), blk_hostobj.__class__.__name__,
      error_lineno ) )

class InvalidIndexError( Exception ):
  """ Raise when a variable in an update block is not declared """
  def __init__( self, obj, idx, blk=None, blk_hostobj=None, lineno=0 ):

    if not blk:
      return super().__init__() # this is just temporary message

    filepath = inspect.getfile( blk_hostobj.__class__ )
    blk_src, base_lineno  = inspect.getsourcelines( blk )

    # Shunning: we need to subtract 1 from inspect's lineno when we add it
    # to base_lineno because it starts from 1!
    lineno -= 1
    error_lineno = base_lineno + lineno

    if isinstance( idx, tuple ):
      idx_str = f"[{idx.start}:{idx.stop}]"
    else:
      idx_str = f"[{idx}]"

    from pymtl3.datatypes import Bits
    return super().__init__( \
"""
In file {}:{} in {}

{} {}
^^^ Slice {} of object \"{}\" (class \"{}\") is accessed in block \"{}\",
    but {} has \"{}\" type{}.
(when constructing instance {} of class \"{}\" in the hierarchy)

Suggestion: fix incorrect field access at line {}, or fix the declaration somewhere.""".format( \
      filepath, error_lineno, blk.__name__,
      error_lineno, blk_src[ lineno ].lstrip(''),
      idx_str, repr(obj), obj.__class__.__name__, blk.__name__,
      repr(obj), obj._dsl.Type.__name__, "" if issubclass( obj._dsl.Type, Bits ) else ", not BitsN type",
      repr(blk_hostobj), blk_hostobj.__class__.__name__,
      error_lineno ) )

class UpblkFuncSameNameError( Exception ):
  """ Raise when two update block/function are declared with the same name """
  def __init__( self, name ):
    return super().__init__( \
      " Cannot declare two update blocks/functions with the same name {}".format( name ) )

class UpblkCyclicError( Exception ):
  """ Raise when update blocks have cyclic dependencies """

class InvalidConstraintError( Exception ):
  """ Raise when a defined constraint is of wrong format """
  def __init__( self ):
    return super().__init__( \
      "Constraints between two variables are not allowed!" )

class InvalidConnectionError( Exception ):
  """ Raise upon an inappropriate attempt to connect two variables """

class InvalidFuncCallError( Exception ):
  """ Raise upon an inappropriate @s.func function call """

class InvalidPlaceholderError( Exception ):
  """ Raise upon declaring an update block in a placeholder component. """

class NotElaboratedError( Exception ):
  """ Raise when processing a model that hasn't been elaborated yet """
  def __init__( self ):
    return super().__init__( \
    "Please elaborate the model first." )

class InvalidAPICallError( Exception ):
  """ Raise when processing a model that hasn't been elaborated yet """
  def __init__( self, api_name, obj, top ):
    return super().__init__( \
    "{} is only allowed to be called at the top module that .elaborate() "
    "was called on (an instance of {}), but this API call is on {}." \
    .format( api_name, top.__class__, "top."+repr(obj)[2:] ) )

class LeftoverPlaceholderError( Exception ):
  """ Raise upon declaring an update block in a placeholder component. """
  def __init__( self, placeholders ):
    return super().__init__( \
    "Please replace all placeholders with valid components:\n - {}".format(
      "\n - ".join( [ "top.{} (instance of {})".format( repr(x)[2:], x.__class__ ) \
                     for x in placeholders ] ) ) )
